# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import math
import pickle

import numpy as np
import pytest
from affine import Affine
from pytest import approx

from odc.geo._interop import have
from odc.geo import CRS, CRSMismatchError, geom, wh_
from odc.geo.geobox import GeoBox, _round_to_res
from odc.geo.geom import (
    chop_along_antimeridian,
    clip_lon180,
    densify,
    force_2d,
    multigeom,
    projected_lon,
    triangulate,
)
from odc.geo.testutils import (
    SAMPLE_WKT_WITHOUT_AUTHORITY,
    epsg3577,
    epsg3857,
    epsg4326,
    from_fixed_point,
    gen_test_image_xy,
    to_fixed_point,
    xy_from_gbox,
)

# pylint: disable=protected-access, pointless-statement
# pylint: disable=too-many-statements,too-many-locals,too-many-lines,unnecessary-lambda-assignment


def test_pickleable():
    poly = geom.polygon([(10, 20), (20, 20), (20, 10), (10, 20)], crs=epsg4326)
    pickled = pickle.dumps(poly, pickle.HIGHEST_PROTOCOL)
    unpickled = pickle.loads(pickled)
    assert poly == unpickled


def test_props():
    crs = epsg4326

    box1 = geom.box(10, 10, 30, 30, crs=crs)
    assert box1
    assert box1.is_valid
    assert not box1.is_empty
    assert box1.area == 400.0
    assert box1.boundary.length == 80.0
    assert box1.centroid == geom.point(20, 20, crs)
    assert isinstance(box1.wkt, str)

    triangle = geom.polygon([(10, 20), (20, 20), (20, 10), (10, 20)], crs=crs)
    assert triangle.boundingbox == geom.BoundingBox(10, 10, 20, 20, crs)
    assert triangle.envelope.contains(triangle)

    assert box1.length == 80.0

    box1copy = geom.box(10, 10, 30, 30, crs=crs)
    assert box1 == box1copy
    # NOTE: this might fail because of point order
    assert box1.convex_hull == box1copy

    box2 = geom.box(20, 10, 40, 30, crs=crs)
    assert box1 != box2

    bbox = geom.BoundingBox(1, 0, 10, 13)
    assert bbox.width == 9
    assert bbox.height == 13
    assert bbox.points == [(1, 0), (1, 13), (10, 0), (10, 13)]

    assert bbox.transform(Affine.identity()) == bbox
    assert bbox.transform(Affine.translation(1, 2)) == geom.BoundingBox(2, 2, 11, 15)

    pt = geom.point(3, 4, crs)
    assert pt.json["coordinates"] == (3.0, 4.0)
    assert "Point" in str(pt)
    assert bool(pt) is True
    assert pt.__nonzero__() is True

    # check "CRS as string is converted to class automatically"
    assert isinstance(geom.point(3, 4, "epsg:3857").crs, geom.CRS)

    # constructor with bad input should raise ValueError
    with pytest.raises(ValueError):
        geom.Geometry(object())


def test_tests():
    box1 = geom.box(10, 10, 30, 30, crs=epsg4326)
    box2 = geom.box(20, 10, 40, 30, crs=epsg4326)
    box3 = geom.box(30, 10, 50, 30, crs=epsg4326)
    box4 = geom.box(40, 10, 60, 30, crs=epsg4326)
    minibox = geom.box(15, 15, 25, 25, crs=epsg4326)

    assert not box1.touches(box2)
    assert box1.touches(box3)
    assert not box1.touches(box4)

    assert box1.intersects(box2)
    assert box1.intersects(box3)
    assert not box1.intersects(box4)

    assert not box1.crosses(box2)
    assert not box1.crosses(box3)
    assert not box1.crosses(box4)

    assert not box1.disjoint(box2)
    assert not box1.disjoint(box3)
    assert box1.disjoint(box4)

    assert box1.contains(minibox)
    assert not box1.contains(box2)
    assert not box1.contains(box3)
    assert not box1.contains(box4)

    assert minibox.within(box1)
    assert not box1.within(box2)
    assert not box1.within(box3)
    assert not box1.within(box4)


def test_ops():
    box1 = geom.box(10, 10, 30, 30, crs=epsg4326)
    box2 = geom.box(20, 10, 40, 30, crs=epsg4326)
    box3 = geom.box(20, 10, 40, 30, crs=epsg4326)
    box4 = geom.box(40, 10, 60, 30, crs=epsg4326)
    no_box = None

    assert box1 != box2
    assert box2 == box3
    assert box3 != no_box

    union1 = box1.union(box2)
    assert union1.area == 600.0

    assert box1.crs == epsg4326
    assert box1.assign_crs(None).crs is None
    assert box1.assign_crs(None).geom is box1.geom

    with pytest.raises(geom.CRSMismatchError):
        box1.union(box2.to_crs(epsg3857))

    inter1 = box1.intersection(box2)
    assert bool(inter1)
    assert inter1.area == 200.0

    inter2 = box1.intersection(box4)
    assert not bool(inter2)
    assert inter2.is_empty
    # assert not inter2.is_valid  TODO: what's going on here?

    diff1 = box1.difference(box2)
    assert diff1.area == 200.0

    symdiff1 = box1.symmetric_difference(box2)
    assert symdiff1.area == 400.0

    # test segmented
    line = geom.line([(0, 0), (0, 5), (10, 5)], epsg4326)
    line2 = line.segmented(2)
    assert line.crs is line2.crs
    assert line.length == line2.length
    assert len(line.coords) < len(line2.coords)
    poly = geom.polygon([(0, 0), (0, 5), (10, 5)], epsg4326)
    poly2 = poly.segmented(2)
    assert poly.crs is poly2.crs
    assert poly.length == poly2.length
    assert poly.area == poly2.area
    assert len(poly.geom.exterior.coords) < len(poly2.geom.exterior.coords)

    poly2 = poly.exterior.segmented(2)
    assert poly.crs is poly2.crs
    assert poly.length == poly2.length
    assert len(poly.geom.exterior.coords) < len(poly2.geom.coords)

    # test point.segmented is just a clone
    pt = geom.point(10, 20, "EPSG:4326")
    assert pt.segmented(1) is not pt
    assert pt.segmented(1) == pt

    # test interpolate
    pt = line.interpolate(1)
    assert pt.crs is line.crs
    assert pt.coords[0] == (0, 1)
    assert isinstance(pt.coords, list)

    with pytest.raises(TypeError):
        pt.interpolate(3)

    # test simplify
    poly = geom.polygon([(0, 0), (0, 5), (10, 5)], epsg4326)
    assert poly.simplify(100) == poly

    # test iteration
    poly_2_parts = geom.Geometry(
        {
            "type": "MultiPolygon",
            "coordinates": [
                [
                    [
                        [102.0, 2.0],
                        [103.0, 2.0],
                        [103.0, 3.0],
                        [102.0, 3.0],
                        [102.0, 2.0],
                    ]
                ],
                [
                    [
                        [100.0, 0.0],
                        [101.0, 0.0],
                        [101.0, 1.0],
                        [100.0, 1.0],
                        [100.0, 0.0],
                    ],
                    [
                        [100.2, 0.2],
                        [100.8, 0.2],
                        [100.8, 0.8],
                        [100.2, 0.8],
                        [100.2, 0.2],
                    ],
                ],
            ],
        },
        "EPSG:4326",
    )
    pp = list(poly_2_parts.geoms)
    assert len(pp) == 2
    assert all(p.crs == poly_2_parts.crs for p in pp)

    # test transform
    assert geom.point(0, 0, epsg4326).transform(
        lambda x, y: (x + 1, y + 2)
    ) == geom.point(1, 2, epsg4326)

    # test sides
    box = geom.box(1, 2, 11, 22, epsg4326)
    lines = list(geom.sides(box))
    assert all(line.crs is epsg4326 for line in lines)
    assert len(lines) == 4
    assert lines[0] == geom.line([(1, 2), (1, 22)], epsg4326)
    assert lines[1] == geom.line([(1, 22), (11, 22)], epsg4326)
    assert lines[2] == geom.line([(11, 22), (11, 2)], epsg4326)
    assert lines[3] == geom.line([(11, 2), (1, 2)], epsg4326)


def test_geom_split():
    box = geom.box(0, 0, 10, 30, epsg4326)
    line = geom.line([(5, 0), (5, 30)], epsg4326)
    bb = list(box.split(line))
    assert len(bb) == 2
    assert box.contains(bb[0] | bb[1])
    assert (box ^ (bb[0] | bb[1])).is_empty

    with pytest.raises(CRSMismatchError):
        list(box.split(geom.line([(5, 0), (5, 30)], epsg3857)))


def test_multigeom():
    p1, p2 = (0, 0), (1, 2)
    p3, p4 = (3, 4), (5, 6)
    b1 = geom.box(*p1, *p2, epsg4326)
    b2 = geom.box(*p3, *p4, epsg4326)
    bb = multigeom([b1, b2])
    assert bb.geom_type == "MultiPolygon"
    assert bb.crs is b1.crs
    assert len(list(bb.geoms)) == 2

    assert geom.multipolygon([[b1.boundary.coords], [b2.boundary.coords]], b1.crs) == bb

    g1 = geom.line([p1, p2], None)
    g2 = geom.line([p3, p4], None)
    gg = multigeom(iter([g1, g2, g1]))
    assert gg.geom_type == "MultiLineString"
    assert gg.crs is g1.crs
    assert len(list(gg.geoms)) == 3
    assert geom.multiline([[p1, p2], [p3, p4], [p1, p2]], None) == gg

    g1 = geom.point(*p1, epsg3857)
    g2 = geom.point(*p2, epsg3857)
    g3 = geom.point(*p3, epsg3857)
    gg = multigeom(iter([g1, g2, g3]))
    assert gg.geom_type == "MultiPoint"
    assert gg.crs is g1.crs
    assert len(list(gg.geoms)) == 3
    assert list(gg.geoms)[0] == g1
    assert list(gg.geoms)[1] == g2
    assert list(gg.geoms)[2] == g3

    assert geom.multipoint([p1, p2, p3], epsg3857) == gg

    # check mixed types
    assert (
        multigeom([geom.line([p1, p2], None), geom.point(*p1, None)]).geom_type
        == "GeometryCollection"
    )

    # can't mix CRSs
    with pytest.raises(CRSMismatchError):
        multigeom([geom.line([p1, p2], epsg4326), geom.line([p3, p4], epsg3857)])

    assert multigeom([gg]).geom_type == "GeometryCollection"


def test_shapely_wrappers():
    poly = geom.polygon([(0, 0), (0, 5), (10, 5)], epsg4326)

    assert isinstance(poly.svg(), str)
    assert isinstance(poly._repr_svg_(), str)

    with_hole = poly.buffer(1) - poly
    assert len(poly.interiors) == 0
    assert len(with_hole.interiors) == 1
    assert isinstance(with_hole.interiors, list)
    assert isinstance(with_hole.interiors[0], geom.Geometry)
    assert with_hole.interiors[0].crs == with_hole.crs
    assert poly.exterior.crs == poly.crs

    x, y = poly.exterior.xy
    assert len(x) == len(y)
    assert x.typecode == y.typecode
    assert x.typecode == "d"

    assert ((poly | poly) - poly).is_empty
    assert ((poly & poly) - poly).is_empty
    assert (poly ^ poly).is_empty
    assert (poly - poly).is_empty


def test_to_crs():
    poly = geom.polygon([(0, 0), (0, 5), (10, 5)], epsg4326)
    src_num_points = len(poly.exterior.xy[0])

    assert poly.crs is epsg4326
    assert poly.to_crs(epsg3857).crs is epsg3857
    assert poly.to_crs("EPSG:3857").crs == "EPSG:3857"
    assert poly.to_crs("EPSG:3857", 0.1).crs == epsg3857
    assert poly.to_crs(3857, "auto").crs == epsg3857

    assert poly.exterior.to_crs(epsg3857) == poly.to_crs(epsg3857).exterior

    # test that by default segmentation happens
    assert len(poly.to_crs(epsg3857, resolution=1).exterior.xy[0]) > src_num_points

    # test no segmentation by default or when inf/nan is supplied
    assert len(poly.to_crs(epsg3857).exterior.xy[0]) == src_num_points
    assert len(poly.to_crs(epsg3857, float("+inf")).exterior.xy[0]) == src_num_points
    assert len(poly.to_crs(epsg3857, float("nan")).exterior.xy[0]) == src_num_points

    # test the segmentation works on multi-polygons
    mpoly = geom.box(0, 0, 1, 3, "EPSG:4326") | geom.box(2, 4, 3, 6, "EPSG:4326")

    assert mpoly.geom_type == "MultiPolygon"
    assert mpoly.to_crs(epsg3857).geom_type == "MultiPolygon"
    assert mpoly.to_crs(epsg3857, 1).geom_type == "MultiPolygon"

    poly = geom.polygon([(0, 0), (0, 5), (10, 5)], None)
    assert poly.crs is None
    with pytest.raises(ValueError):
        poly.to_crs(epsg3857)


def test_to_crs_with_check():
    crs = "+proj=ortho +lat0=40"
    gg = geom.line([[-180, 0], [180, 0]], 4326).segmented(5)
    assert gg.to_crs(crs).is_valid is False
    assert gg.to_crs(crs, check_and_fix=True).is_valid is True

    gg = gg.buffer(1.1) & geom.box(-180, -90, 180, 90, 4326)

    assert gg.to_crs(crs).is_valid is False
    assert gg.to_crs(crs, check_and_fix=True).is_valid is True


def test_to_crs_utm():
    poly = geom.box(0.1, 43, 1.3, 44, epsg4326)
    assert poly.to_crs("utm").crs.epsg == 32631
    assert poly.to_crs("utm") == poly.to_crs(32631)
    assert poly.to_crs("utm-s").crs.epsg == 32731

    poly = geom.box(0.1, -44, 1.3, -43, epsg4326)
    assert poly.to_crs("utm-n").crs.epsg == 32631
    assert poly.to_crs("utm") == poly.to_crs(32731)


def test_densify():
    s_x10 = [(0, 0), (10, 0)]
    assert densify(s_x10, 20) == s_x10
    assert densify(s_x10, 200) == s_x10
    assert densify(s_x10, 5) == [(0, 0), (5, 0), (10, 0)]
    assert densify(s_x10, 4) == [(0, 0), (4, 0), (8, 0), (10, 0)]


def test_unary_union():
    box1 = geom.box(10, 10, 30, 30, crs=epsg4326)
    box2 = geom.box(20, 10, 40, 30, crs=epsg4326)
    box3 = geom.box(30, 10, 50, 30, crs=epsg4326)
    box4 = geom.box(40, 10, 60, 30, crs=epsg4326)

    union0 = geom.unary_union([box1])
    assert union0 == box1

    union1 = geom.unary_union([box1, box4])
    assert union1.geom_type == "MultiPolygon"
    assert union1.area == 2.0 * box1.area

    union2 = geom.unary_union([box1, box2])
    assert union2.geom_type == "Polygon"
    assert union2.area == 1.5 * box1.area

    union3 = geom.unary_union([box1, box2, box3, box4])
    assert union3.geom_type == "Polygon"
    assert union3.area == 2.5 * box1.area

    union4 = geom.unary_union([union1, box2, box3])
    assert union4.geom_type == "Polygon"
    assert union4.area == 2.5 * box1.area

    assert geom.unary_union([]) is None

    with pytest.raises(ValueError):
        geom.unary_union([box1, box1.to_crs(epsg3577)])


def test_unary_intersection():
    box1 = geom.box(10, 10, 30, 30, crs=epsg4326)
    box2 = geom.box(15, 10, 35, 30, crs=epsg4326)
    box3 = geom.box(20, 10, 40, 30, crs=epsg4326)
    box4 = geom.box(25, 10, 45, 30, crs=epsg4326)
    box5 = geom.box(30, 10, 50, 30, crs=epsg4326)
    box6 = geom.box(35, 10, 55, 30, crs=epsg4326)

    assert geom.intersects(box1, box2)
    inter1 = geom.unary_intersection([box1])
    assert bool(inter1)
    assert inter1 == box1

    inter2 = geom.unary_intersection([box1, box2])
    assert bool(inter2)
    assert inter2.area == 300.0

    inter3 = geom.unary_intersection([box1, box2, box3])
    assert bool(inter3)
    assert inter3.area == 200.0

    inter4 = geom.unary_intersection([box1, box2, box3, box4])
    assert bool(inter4)
    assert inter4.area == 100.0

    inter5 = geom.unary_intersection([box1, box2, box3, box4, box5])
    assert bool(inter5)
    assert inter5.geom_type == "LineString"
    assert inter5.length == 20.0

    inter6 = geom.unary_intersection([box1, box2, box3, box4, box5, box6])
    assert not bool(inter6)
    assert inter6.is_empty


def test_gen_test_image_xy():
    gbox = GeoBox(wh_(3, 7), Affine.translation(10, 1000), epsg3857)

    xy, denorm = gen_test_image_xy(gbox, "float64")
    assert xy.dtype == "float64"
    assert xy.shape == (2,) + gbox.shape

    x, y = denorm(xy)
    x_, y_ = xy_from_gbox(gbox)

    np.testing.assert_almost_equal(x, x_)
    np.testing.assert_almost_equal(y, y_)

    xy, denorm = gen_test_image_xy(gbox, "uint16")
    assert xy.dtype == "uint16"
    assert xy.shape == (2,) + gbox.shape

    x, y = denorm(xy[0], xy[1])
    assert x.shape == xy.shape[1:]
    assert y.shape == xy.shape[1:]
    assert x.dtype == "float64"

    x_, y_ = xy_from_gbox(gbox)

    np.testing.assert_almost_equal(x, x_, 4)
    np.testing.assert_almost_equal(y, y_, 4)

    for dt in ("int8", np.int16, np.dtype(np.uint64)):
        xy, _ = gen_test_image_xy(gbox, dt)
        assert xy.dtype == dt

    # check no-data
    xy, denorm = gen_test_image_xy(gbox, "float32")
    assert xy.dtype == "float32"
    assert xy.shape == (2,) + gbox.shape
    xy[0, 0, :] = np.nan
    xy[1, 1, :] = np.nan
    _xy = denorm(xy, nodata=np.nan)
    assert np.isnan(_xy[:, :2]).all()
    np.testing.assert_almost_equal(_xy[0][2:], x_[2:], 6)
    np.testing.assert_almost_equal(_xy[1][2:], y_[2:], 6)

    xy, denorm = gen_test_image_xy(gbox, "int16")
    assert xy.dtype == "int16"
    assert xy.shape == (2,) + gbox.shape
    xy[0, 0, :] = -999
    xy[1, 1, :] = -999
    _xy = denorm(xy, nodata=-999)
    assert np.isnan(_xy[:, :2]).all()
    np.testing.assert_almost_equal(_xy[0][2:], x_[2:], 4)
    np.testing.assert_almost_equal(_xy[1][2:], y_[2:], 4)

    # call without arguments should return linear mapping
    A = denorm()
    assert isinstance(A, Affine)


def test_fixed_point():
    aa = np.asarray([0, 0.5, 1])
    uu = to_fixed_point(aa, "uint8")
    assert uu.dtype == "uint8"
    assert aa.shape == uu.shape
    assert tuple(uu.ravel()) == (0, 128, 255)

    aa_ = from_fixed_point(uu)
    assert aa_.dtype == "float64"
    dd = np.abs(aa - aa_)
    assert (dd < 1 / 255.0).all()

    uu = to_fixed_point(aa, "uint16")
    assert uu.dtype == "uint16"
    assert tuple(uu.ravel()) == (0, 0x8000, 0xFFFF)

    uu = to_fixed_point(aa, "int16")
    assert uu.dtype == "int16"
    assert tuple(uu.ravel()) == (0, 0x4000, 0x7FFF)

    aa_ = from_fixed_point(uu)
    dd = np.abs(aa - aa_)
    assert (dd < 1.0 / 0x7FFF).all()


def test_projected_lon():
    assert projected_lon(epsg3857, 180).crs is epsg3857
    assert projected_lon("EPSG:3577", 100).crs == epsg3577


def test_chop():
    poly = geom.box(618300, -1876800, 849000, -1642500, "EPSG:32660")

    chopped = chop_along_antimeridian(poly)
    assert chopped.crs is poly.crs
    assert chopped.geom_type == "MultiPolygon"
    assert len(list(chopped.geoms)) == 2

    poly = geom.box(0, 0, 10, 20, "EPSG:4326")._to_crs(epsg3857)
    assert poly.crs is epsg3857
    assert chop_along_antimeridian(poly) is poly

    with pytest.raises(ValueError):
        chop_along_antimeridian(geom.box(0, 1, 2, 3, None))


def test_clip_lon180():
    err = 1e-9

    def b(rside):
        return geom.box(170, 0, rside, 10, epsg4326)

    def b_neg(lside):
        return geom.box(lside, 0, -170, 10, epsg4326)

    assert clip_lon180(b(180 - err)) == b(180)
    assert clip_lon180(b(-180 + err)) == b(180)

    assert clip_lon180(b_neg(180 - err)) == b_neg(-180)
    assert clip_lon180(b_neg(-180 + err)) == b_neg(-180)

    bb = multigeom([b(180 - err), b_neg(180 - err)])
    bb_ = list(clip_lon180(bb).geoms)
    assert bb_[0] == b(180)
    assert bb_[1] == b_neg(-180)


def test_wrap_dateline():
    albers_crs = epsg3577
    geog_crs = epsg4326

    wrap = geom.polygon(
        [
            (3658653.1976781483, -4995675.379595791),
            (4025493.916030875, -3947239.249752495),
            (4912789.243100313, -4297237.125269571),
            (4465089.861944263, -5313778.16975072),
            (3658653.1976781483, -4995675.379595791),
        ],
        crs=albers_crs,
    )
    wrapped = wrap.to_crs(geog_crs)
    assert wrapped.geom_type == "Polygon"
    assert wrapped.intersects(geom.line([(0, -90), (0, 90)], crs=geog_crs))
    wrapped = wrap.to_crs(geog_crs, wrapdateline=True)
    assert wrapped.geom_type == "MultiPolygon"
    assert not wrapped.intersects(geom.line([(0, -90), (0, 90)], crs=geog_crs))


@pytest.mark.parametrize(
    "pts",
    [
        [
            (12231455.716333, -5559752.598333),
            (12231455.716333, -4447802.078667),
            (13343406.236, -4447802.078667),
            (13343406.236, -5559752.598333),
            (12231455.716333, -5559752.598333),
        ],
        [
            (13343406.236, -5559752.598333),
            (13343406.236, -4447802.078667),
            (14455356.755667, -4447802.078667),
            (14455356.755667, -5559752.598333),
            (13343406.236, -5559752.598333),
        ],
        [
            (14455356.755667, -5559752.598333),
            (14455356.755667, -4447802.078667),
            (15567307.275333, -4447802.078667),
            (15567307.275333, -5559752.598333),
            (14455356.755667, -5559752.598333),
        ],
    ],
)
def test_wrap_dateline_sinusoidal(pts):
    sinus_crs = geom.CRS(
        """PROJCS["unnamed",
                           GEOGCS["Unknown datum based upon the custom spheroid",
                           DATUM["Not specified (based on custom spheroid)", SPHEROID["Custom spheroid",6371007.181,0]],
                           PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],
                           PROJECTION["Sinusoidal"],
                           PARAMETER["longitude_of_center",0],
                           PARAMETER["false_easting",0],
                           PARAMETER["false_northing",0],
                           UNIT["Meter",1]]"""
    )

    wrap = geom.polygon(pts, crs=sinus_crs)
    wrapped = wrap.to_crs(epsg4326)
    assert wrapped.geom_type == "Polygon"
    wrapped = wrap.to_crs(epsg4326, wrapdateline=True)
    assert wrapped.geom_type == "MultiPolygon"
    assert not wrapped.intersects(geom.line([(0, -90), (0, 90)], crs=epsg4326))


def test_wrap_dateline_utm():
    poly = geom.box(618300, -1876800, 849000, -1642500, "EPSG:32660")

    wrapped = poly.to_crs(epsg4326)
    assert wrapped.geom_type == "Polygon"
    assert wrapped.intersects(geom.line([(0, -90), (0, 90)], crs=epsg4326))
    wrapped = poly.to_crs(epsg4326, wrapdateline=True)
    assert wrapped.geom_type == "MultiPolygon"
    assert not wrapped.intersects(geom.line([(0, -90), (0, 90)], crs=epsg4326))


def test_3d_geometry_converted_to_2d_geometry():
    coordinates = [
        (115.8929714190001, -28.577007674999948, 0.0),
        (115.90275429200005, -28.57698532699993, 0.0),
        (115.90412631000004, -28.577577566999935, 0.0),
        (115.90157040700001, -28.58521105999995, 0.0),
        (115.89382838900008, -28.585473711999953, 0.0),
        (115.8929714190001, -28.577007674999948, 0.0),
    ]
    geom_3d = {"coordinates": [coordinates], "type": "Polygon"}
    geom_2d = {"coordinates": [[(x, y) for x, y, z in coordinates]], "type": "Polygon"}

    g_2d = geom.Geometry(geom_2d)
    g_3d = geom.Geometry(geom_3d)

    assert {2} == {len(pt) for pt in g_3d.boundary.coords}  # All coordinates are 2D

    assert g_2d == g_3d  # 3D geometry has been converted to a 2D by dropping the Z axis


def test_3d_point_converted_to_2d_point():
    point = (-35.5029340, 145.9312455, 0.0)

    point_3d = {"coordinates": point, "type": "Point"}
    point_2d = {"coordinates": (point[0], point[1]), "type": "Point"}

    p_2d = geom.Geometry(point_2d)
    p_3d = geom.Geometry(point_3d)

    assert len(p_3d.coords[0]) == 2

    assert p_2d == p_3d


def test_mid_longitude():
    assert geom.mid_longitude(geom.point(10, 3, "epsg:4326")) == approx(10)
    assert geom.mid_longitude(geom.point(10, 3, "epsg:4326").buffer(3)) == approx(10)
    assert geom.mid_longitude(
        geom.point(10, 3, "epsg:4326").buffer(3).to_crs("epsg:3857")
    ) == approx(10)


def test_point_transformer():
    tr = epsg3857.transformer_to_crs(epsg4326)
    tr_back = epsg4326.transformer_to_crs(epsg3857)

    pts = [(0, 0), (0, 1), (1, 2), (10, 11)]
    x, y = np.vstack(pts).astype("float64").T

    pts_expect = [geom.point(*pt, epsg3857).to_crs(epsg4326).points[0] for pt in pts]

    x_expect = [pt[0] for pt in pts_expect]
    y_expect = [pt[1] for pt in pts_expect]

    x_, y_ = tr(x, y)
    assert x_.shape == x.shape
    np.testing.assert_array_almost_equal(x_, x_expect)
    np.testing.assert_array_almost_equal(y_, y_expect)

    x, y = (a.reshape(2, 2) for a in (x, y))
    x_, y_ = tr(x, y)
    assert x_.shape == x.shape

    xb, yb = tr_back(x_, y_)
    np.testing.assert_array_almost_equal(x, xb)
    np.testing.assert_array_almost_equal(y, yb)

    # check nans
    x_, y_ = tr(np.asarray([np.nan, 0, np.nan]), np.asarray([0, np.nan, np.nan]))

    assert np.isnan(x_).all()
    assert np.isnan(y_).all()


def test_base_internals():
    no_epsg_crs = CRS(SAMPLE_WKT_WITHOUT_AUTHORITY)
    assert no_epsg_crs.epsg is None

    gjson_bad = {"type": "a", "coordinates": [1, [2, 3, 4]]}
    assert force_2d(gjson_bad) == {"type": "a", "coordinates": [1, [2, 3]]}

    with pytest.raises(ValueError):
        force_2d({"type": "a", "coordinates": [set("not a valid element")]})

    assert _round_to_res(0.2, 1.0) == 1
    assert _round_to_res(0.0, 1.0) == 0
    assert _round_to_res(0.05, 1.0) == 0


def test_geom_clone():
    b = geom.box(0, 0, 10, 20, epsg4326)
    assert b == b.clone()
    assert b.geom is not b.clone().geom

    assert b == geom.Geometry(b)
    assert b.geom is not geom.Geometry(b).geom


@pytest.mark.parametrize("crs", [None, epsg4326])
@pytest.mark.parametrize("nside", [3, 7, 10])
def test_bbox_boundary(crs, nside):
    n_pts_expect = (nside - 1) * 4
    bbox = geom.BoundingBox(0, 20, 11, 29, crs)
    poly = bbox.polygon
    bb = bbox.boundary()
    assert bb.crs == bbox.crs
    assert bb.is_ring
    assert bb.geom_type == "LineString"
    assert len(bb.coords[:-1]) == 4
    assert (poly.exterior ^ bbox.boundary()).is_empty
    assert (bbox.boundary() - poly).is_empty

    bb = bbox.boundary(nside)
    assert n_pts_expect == len(bb.coords[:-1])
    assert (bbox.boundary() - poly).is_empty


def test_lonlat_bounds():
    # example from landsat scene: spans lon=180
    poly = geom.box(618300, -1876800, 849000, -1642500, "EPSG:32660")

    bb = geom.lonlat_bounds(poly)
    assert bb.left < 180 < bb.right
    assert geom.lonlat_bounds(poly) == geom.lonlat_bounds(poly, resolution=1e8)

    bb = geom.lonlat_bounds(poly, mode="quick")
    assert bb.right - bb.left > 180

    poly = geom.box(1, -10, 2, 20, "EPSG:4326")
    assert geom.lonlat_bounds(poly) == poly.boundingbox

    with pytest.raises(ValueError):
        geom.lonlat_bounds(geom.box(0, 0, 1, 1, None))

    multi = {
        "type": "MultiPolygon",
        "coordinates": [
            [[[174, 52], [174, 53], [175, 53], [174, 52]]],
            [[[168, 54], [167, 55], [167, 54], [168, 54]]],
        ],
    }

    multi_geom = geom.Geometry(multi, "epsg:4326")
    multi_geom_projected = multi_geom.to_crs("epsg:32659")
    expect = approx(geom.lonlat_bounds(multi_geom))

    assert geom.lonlat_bounds(multi_geom_projected) == expect
    assert geom.lonlat_bounds(multi_geom_projected, resolution="auto") == expect


def test_geojson():
    b = geom.box(0, 0, 10, 20, epsg4326)
    gjson = b.geojson()
    assert set(list(gjson)) == set(["type", "geometry", "properties"])
    assert gjson["type"] == "Feature"
    assert gjson["properties"] == {}
    _b = geom.Geometry(gjson["geometry"], crs=epsg4326)
    assert (b - _b).area < 1e-6

    # crs=None case should work too
    gjson = b.assign_crs(None).geojson()
    assert set(list(gjson)) == set(["type", "geometry", "properties"])
    assert gjson["type"] == "Feature"
    assert gjson["properties"] == {}
    _b = geom.Geometry(gjson["geometry"], crs=epsg4326)
    assert (b - _b).area < 1e-6

    gjson = b.to_crs(epsg3857).geojson(region_code="33")
    assert set(list(gjson)) == set(["type", "geometry", "properties"])
    assert gjson["type"] == "Feature"
    assert gjson["properties"] == {"region_code": "33"}

    _b = geom.Geometry(gjson, crs=epsg4326)
    assert (b - _b).area < 1e-6

    _b = geom.Geometry(gjson)
    assert _b.crs == epsg4326
    assert (b - _b).area < 1e-6

    _b = geom.Geometry(dict(type="FeatureCollection", features=[gjson]))
    assert _b.crs == epsg4326
    assert (b - _b).area < 1e-6

    _b = geom.Geometry(gjson["geometry"], crs=epsg4326)
    assert (b - _b).area < 1e-6

    with pytest.raises(ValueError):
        _ = geom.Geometry({})

    # check FeatureCollection output
    gg = b.buffer(0.1).exterior | b.centroid
    assert gg.geom_type == "GeometryCollection"
    gjson = gg.geojson(k=3)
    assert gjson["type"] == "FeatureCollection"
    assert set(gjson) == set(["type", "features"])
    assert gjson["features"][0]["properties"]["k"] == 3
    assert gjson["features"][1]["properties"]["k"] == 3
    gg_ = geom.Geometry(gjson)
    assert gg_.geom_type == "GeometryCollection"
    assert gg_.crs == "epsg:4326"
    assert len(list(gg_.geoms)) == len(list(gg.geoms))

    gg = geom.Geometry(
        {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {"type": "Point", "coordinates": [0, 1]},
                },
                {
                    "type": "Feature",
                    "geometry": {"type": "Point", "coordinates": [10, 20]},
                },
            ],
        }
    )
    assert gg.geom_type == "MultiPoint"
    assert len(list(gg.geoms)) == 2


@pytest.mark.xfail(
    True, reason="Bounds computation for large geometries in safe mode is broken"
)
def test_lonalt_bounds_more_than_180():
    poly = geom.box(-150, -30, 150, 30, epsg4326).to_crs(epsg3857, math.inf)

    assert geom.lonlat_bounds(poly, "quick") == approx((-150, -30, 150, 30))
    assert geom.lonlat_bounds(poly, "safe") == approx((-150, -30, 150, 30))


def test_mul_affine():
    g = geom.point(1, 2, epsg4326)
    x10 = Affine.scale(10)

    with pytest.warns(DeprecationWarning):
        # it warns because x10.__mul__(g) is called first and it attempts to
        # iterate over g, which is deprecated.
        #
        # Only then g.__rmul__(x10) is called
        assert Affine.translation(10, 100) * g == geom.point(11, 102, epsg4326)

        assert x10 * g == geom.point(10, 20, epsg4326)
        assert x10 * g == g.transform(x10)

    mg = geom.multigeom([g, g, g])
    assert mg.is_multi
    assert mg.transform(x10).is_multi
    assert mg.transform(x10).crs == mg.crs
    assert mg.transform(x10, crs=None).crs is None

    with pytest.warns(DeprecationWarning):
        assert x10 * mg == mg.transform(x10)


def test_deprecation_warnings():
    g = geom.point(1, 2, epsg4326)
    mg = g | g.buffer(10).boundary
    assert mg.is_multi

    with pytest.warns(DeprecationWarning):
        for _ in mg:
            pass


def test_filter():
    def _keep(x, y):
        return True

    def _drop(x, y):
        return False

    pt = geom.point(1, 2, epsg4326)
    poly = geom.box(1, 2, 3, 4, epsg4326)
    ring = poly.exterior
    multi_pt = geom.multipoint([(1, 1), (2, 2)], epsg4326)
    poly_with_holes = poly - poly.centroid.buffer(0.1)
    assert len(poly_with_holes.interiors) > 0

    for g in [
        pt,
        poly,
        poly_with_holes,
        ring,
        multi_pt,
        pt | ring,
        ring.buffer(1) | ring,
    ]:
        assert g.filter(_drop).is_empty
        assert g.filter(_keep) == g
        assert g.dropna() == g

    multi_pt_nan = multi_pt | geom.point(float("nan"), 7, epsg4326)
    assert len(list(multi_pt_nan.geoms)) == len(list(multi_pt.geoms)) + 1
    assert multi_pt_nan.dropna() == multi_pt


@pytest.mark.parametrize("crs", [None, "epsg:4326"])
@pytest.mark.parametrize("n", [10, 100, 23])
@pytest.mark.parametrize("with_edges", [False, True])
def test_qr2sample(crs, n, with_edges):
    bbox = geom.BoundingBox(10, 33, 21, 55, crs)

    def run_checks(g: geom.Geometry):
        assert g.crs == bbox.crs
        assert (g - bbox.polygon).is_empty
        assert g.is_multi
        assert g.geom_type == "MultiPoint"
        if with_edges is False:
            assert len(list(g.geoms)) == n
        else:
            assert len(list(g.geoms)) > n
            assert (bbox.polygon - triangulate(g)).is_empty

    xx1 = bbox.qr2sample(n, with_edges=with_edges)
    xx2 = bbox.qr2sample(n, offset=100, with_edges=with_edges)
    assert not (xx1 ^ xx2).is_empty
    run_checks(xx1)
    run_checks(xx2)

    xx1_ = bbox.qr2sample(n, with_edges=with_edges)
    assert (xx1 ^ xx1_).is_empty


@pytest.mark.parametrize("crs", [None, "epsg:4326"])
def test_triangulate(crs):
    bbox = geom.BoundingBox(0, 0, 1, 1, crs)

    gg = triangulate(bbox.boundary())
    geoms = list(gg.geoms)
    assert gg.crs == bbox.crs
    assert gg.is_multi
    assert gg.is_valid
    assert len(geoms) == 2
    assert geoms[0].geom_type == "Polygon"
    assert (gg - bbox.polygon).is_empty

    gg = triangulate(bbox.boundary(4), edges=True)
    geoms = list(gg.geoms)
    assert gg.crs == bbox.crs
    assert gg.is_multi
    assert gg.is_valid
    assert geoms[0].geom_type == "LineString"
    assert (gg - bbox.polygon).is_empty


@pytest.mark.skipif(have.folium is False, reason="No folium installed")
@pytest.mark.parametrize(
    "geom_json",
    [
        {"type": "Point", "coordinates": (130, -35)},
        {"type": "LineString", "coordinates": [(110, -40), (150, -30)]},
        {
            "type": "Polygon",
            "coordinates": [
                ((110, -40), (110, -30), (150, -30), (150, -40), (110, -40))
            ],
        },
        {
            "type": "MultiPolygon",
            "coordinates": [
                [[[174, 52], [174, 53], [175, 53], [174, 52]]],
                [[[168, 54], [167, 55], [167, 54], [168, 54]]],
            ],
        },
    ],
)
def test_explore_geom(geom_json):
    from folium import Map, GeoJson

    # Create Geometry
    geometry = geom.Geometry(geom_json, "EPSG:4326")

    # Test explore on dataset input and verify that output is a folium map
    # that contains a GeoJson layer
    m = geometry.explore()
    assert isinstance(m, Map)
    assert any(isinstance(child, GeoJson) for child in m._children.values())

    # Convert to a projected CRS and re-test
    m = geometry.to_crs("EPSG:3577").explore()
    assert isinstance(m, Map)
    assert any(isinstance(child, GeoJson) for child in m._children.values())

    # Verify that passing a custom map works correctly
    m_external = Map()
    geometry.explore(map=m_external)
    assert isinstance(m_external, Map)
    assert any(isinstance(child, GeoJson) for child in m_external._children.values())



from typing import Any, Dict, Optional, Tuple

import xarray as xr

from ._interop import have
from ._rgba import colorize, is_rgb
from .converters import map_crs
from .gcp import GCPGeoBox
from .geobox import GeoBox


# pylint: disable=import-outside-toplevel, redefined-builtin, too-many-locals
def _add_to_folium(url, bounds, map, name=None, index=None, **kw):
    assert have.folium

    from folium.raster_layers import ImageOverlay

    img_overlay = ImageOverlay(url, bounds, name=name, **kw)
    if map is not None:
        img_overlay.add_to(map, name=name, index=index)
    return img_overlay


def _add_to_ipyleaflet(url, bounds, map, name=None, index=None, **kw):
    assert have.ipyleaflet

    from ipyleaflet import ImageOverlay, Map

    assert isinstance(map, Map)

    if name is not None:
        kw.update(name=name)

    img_overlay = ImageOverlay(url=url, bounds=bounds, **kw)
    if map is not None:
        map.add(img_overlay, index=index)

    return img_overlay


def _get_add_to_method(map):
    if map is None:
        return None

    map_module = getattr(map, "__module__", "")

    if "folium" in map_module:
        return _add_to_folium

    if "ipyleaflet" in map_module:
        return _add_to_ipyleaflet

    raise ValueError(f"Not sure how to add image to: '{type(map)}'")


def add_to(
    xx: Any,
    map: Any,
    *,
    name: Optional[str] = None,
    index: Optional[int] = None,
    fmt: str = "png",
    max_size: int = 4096,
    resampling: str = "nearest",
    # jpeg options:
    transparent_pixel: Optional[Tuple[int, int, int]] = None,
    # RGB conversion parameters
    cmap: Optional[Any] = None,
    clip: bool = False,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    robust: Optional[bool] = None,
    # passed to ImageOverlay constructor
    **kw,
) -> Any:
    """
    Add image to an interactive map.

    If map is not supplied, image data url and bounds are returned instead.

    :param xx:
        The :py:class:`~xarray.DataArray` to display
    :param map:
        Map object, :py:mod:`folium` and :py:mod:`ipyleaflet` are
        understood; can also be ``None`` which will return an image data
        url and bounds instead.

    :param name:
        The name of the layer as it will appear in :py:mod:`folium` and
        :py:mod:`ipyleaflet` Layer Controls. The default ``None`` will
        use the input array name (e.g. ``xx.name``) if it exists.
    :param fmt:
        Compress image format. Defaults to "png"; also supports "webp",
        "jpeg".
    :param max_size:
        If longest dimension is bigger than this, shrink it down before
        compression; defaults to 4096.
    :param resampling:
        Custom resampling method to use when reprojecting ``xx`` to the
        map CRS; defaults to "nearest".

    :param transparent_pixel:
        Replace transparent pixels with this value, needed for "jpeg".

    :param cmap:
        If supplied array is not RGB use this colormap to turn it into one.
    :param clip:
        When converting to RGB clip input values to fit ``cmap``.
    :param vmin: Used with matplotlib colormaps
    :param vmax: Used with matplotlib colormaps
    :param robust: Used with matplotlib colormaps, ``vmin=2%, vmax=98%``

    :raises ValueError: when map object is not understood
    :return: ImageLayer that was added to a map
    :return: ``(url, bounds)`` when ``map is None``.

    .. seealso:: :py:meth:`~odc.geo.xr.colorize`, :py:meth:`~odc.geo.xr.to_rgba`

    """
    from .xr import ODCExtensionDa

    assert isinstance(xx, xr.DataArray)
    assert isinstance(xx.odc, ODCExtensionDa)

    _add_to = _get_add_to_method(map)  # raises on error
    _crs = map_crs(map)

    # If array xx has a name (xx.name), use it by default
    if (name is None) and (xx.name is not None):
        name = xx.name if isinstance(xx.name, str) else str(xx.name)

    gbox0 = xx.odc.geobox
    assert gbox0 is not None
    native_crs = gbox0.crs
    assert native_crs is not None
    gbox = gbox0

    if isinstance(gbox, GCPGeoBox):
        if _crs is not None:
            gbox = xx.odc.output_geobox(_crs, tight=True)
        else:
            gbox = xx.odc.output_geobox(native_crs, tight=True)

    if _crs is not None and gbox.crs != _crs:
        gbox = gbox.to_crs(_crs, tight=True)

    if not gbox.axis_aligned:
        gbox = GeoBox.from_bbox(
            gbox.boundingbox, resolution=gbox.resolution, tight=True
        )

    if max(*gbox.shape) > max_size:
        gbox = gbox.zoom_to(max_size)

    if gbox is not gbox0:
        xx = xx.odc.reproject(gbox, resampling=resampling)

    if not is_rgb(xx):
        xx = colorize(xx, cmap=cmap, clip=clip, vmin=vmin, vmax=vmax, robust=robust)

    compress_opts = [fmt]
    for opt in ["zlevel", "quality"]:
        if (v := kw.pop(opt, None)) is not None:
            compress_opts.append(v)

    url = xx.odc.compress(
        *compress_opts, as_data_url=True, transparent=transparent_pixel
    )
    bounds = gbox.map_bounds()
    if _add_to is None:
        return url, bounds

    return _add_to(url, bounds, map, name=name, index=index, **kw)


def explore(
    xx: Any,
    map: Optional[Any] = None,
    *,
    bands: Optional[Tuple[str, str, str]] = None,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    cmap: Optional[Any] = None,
    robust: bool = False,
    tiles: Any = "OpenStreetMap",
    attr: Optional[str] = None,
    layer_control: bool = True,
    resampling: str = "nearest",
    map_kwds: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> Any:
    """
    Plot xarray data on an interactive :py:mod:`folium` leaflet map for
    rapid data exploration.

    :py:class:`xarray.Dataset` inputs are automatically converted to
    multi-band RGB plots, while single-band :py:class:`xarray.DataArray`
    inputs can be plotted using matplotlib colormaps (needs matplotlib
    installed).

    :param xx:
        The :py:class:`~xarray.Dataset` or :py:class:`~xarray.DataArray`
        to plot on the map.
    :param map:
        An optional existing :py:mod:`folium` map object to plot into.
        By default, a new map object will be created.
    :param bands:
        Bands used for RGB colours when converting from a
        :py:class:`~xarray.Dataset` (order should be red, green, blue).
        By default, the function will attempt to guess bands
        automatically. Ignored for :py:class:`~xarray.DataArray` inputs.
    :param vmin:
        Lower value used for the color stretch.
    :param vmax:
        Upper value used for the color stretch.
    :param cmap:
        The colormap used to colorise single-band arrays. If not
        provided, this will default to 'viridis'. Ignored for multi-band
        inputs.
    :param robust:
        If ``True`` (and ``vmin`` and ``vmax`` are absent), the colormap
        range will be computed based on 2nd and 98th percentiles,
        minimising the influence of extreme values. Used for single-band
        arrays only; ignored for multi-band inputs.
    :param tiles:
        Map tileset to use for the map basemap. Supports any option
        supported by :py:mod:`folium`, including "OpenStreetMap",
        "CartoDB positron", "CartoDB dark_matter" or a custom XYZ URL.
    :param attr:
        Map tile attribution; only required if passing custom tile URL.
    :param layer_control:
        Whether to add a control to the map to show or hide map layers.
        If a layer control already exists, this will be skipped.
    :param resampling:
        Custom resampling method to use when reprojecting ``xx`` to the
        map CRS; defaults to "nearest".
    :param map_kwds:
        Additional keyword arguments to pass to :py:class:`folium.Map`.
    :param kwargs:
        Additional keyword arguments to pass to ``.odc.add_to()``.

    :return: A :py:mod:`folium` map containing the plotted xarray data.
    """
    # pylint: disable=too-many-arguments, protected-access

    if not have.folium:
        raise ModuleNotFoundError(
            "'folium' is required but not installed. "
            "Please install it before using `.explore()`."
        )

    from folium import LayerControl, Map

    map_kwds = {} if map_kwds is None else map_kwds
    new_map = map is None

    # If input is a dataset, convert to an RGBA array
    if isinstance(xx, xr.Dataset):
        xx = xx.odc.to_rgba(bands=bands, vmin=vmin, vmax=vmax)

    # Create folium Map if required
    if map is None:
        map = Map(tiles=tiles, attr=attr, **map_kwds)

    # Add to map and raise a friendly error if data has unsuitable dims
    try:
        xx.odc.add_to(
            map,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            robust=robust,
            resampling=resampling,
            **kwargs,
        )
    except ValueError as e:
        raise ValueError(
            "Only 2D single-band (x, y) or 3D multi-band (x, y, band) "
            "arrays are supported by `.explore()`. Please reduce the "
            "dimensions in your array, for example by using `.isel()` "
            "or `.sel()`: `da.isel(time=0).odc.explore()`."
        ) from e

    if new_map:
        # Zoom map to extent of data
        map.fit_bounds(xx.odc.map_bounds())

        # Add a layer control if requested
        if layer_control:
            LayerControl().add_to(map)

    return map



"""
Tools for interop with other libraries.

Check if libraries available without importing them which can be slow.
"""

import importlib.util
from typing import Any, Callable, Optional

# pylint: disable=import-outside-toplevel


class _LibChecker:
    @property
    def rasterio(self) -> bool:
        return self._check("rasterio")

    @property
    def xarray(self) -> bool:
        return self._check("xarray")

    @property
    def geopandas(self) -> bool:
        return self._check("geopandas")

    @property
    def dask(self) -> bool:
        return self._check("dask")

    @property
    def folium(self) -> bool:
        return self._check("folium")

    @property
    def ipyleaflet(self) -> bool:
        return self._check("ipyleaflet")

    @property
    def datacube(self) -> bool:
        return self._check("datacube")

    @property
    def tifffile(self) -> bool:
        return self._check("tifffile")

    @property
    def azure(self) -> bool:
        return self._check("azure.storage.blob")

    @property
    def botocore(self) -> bool:
        return self._check("botocore")

    @staticmethod
    def _check(lib_name: str) -> bool:
        return importlib.util.find_spec(lib_name) is not None

    def check_or_error(self, *libs: str, msg: Optional[str] = None) -> None:
        for lib in libs:
            if not self._check(lib):
                if msg is None:
                    msg = f"Missing library: {lib}"
                raise RuntimeError(msg)


have = _LibChecker()
__all__ = ("have",)


def __dir__():
    return [*__all__, "is_dask_collection"]


def __getattr__(name):
    if name == "is_dask_collection":
        if have.dask:
            import dask

            return dask.is_dask_collection

        # pylint: disable=redefined-outer-name
        def is_dask_collection(_: Any) -> bool:
            return False

        return is_dask_collection
    if name == "__path__":
        loader = globals().get("__loader__")
        if loader is not None:
            return loader.path

    raise AttributeError(f"module {__name__} has no attribute {name}")


is_dask_collection: Callable[[Any], bool]



from collections import abc
from typing import Tuple

import pytest

from odc.geo import ixy_, iyx_, res_, resxy_, resyx_, shape_, wh_, xy_, yx_
from odc.geo.geom import point
from odc.geo.types import func2map

# pylint: disable=use-implicit-booleaness-not-comparison


def test_basics():
    assert xy_(0, 2) == xy_([0, 2])
    assert xy_(0, 2) == xy_(tuple([0, 2]))
    assert yx_(0, 2) == xy_(2, 0)
    assert xy_(1, 0).xy == (1, 0)
    assert xy_(1, 0).lonlat == (1, 0)
    assert xy_(1, 0).yx == (0, 1)
    assert yx_([1, 0]).yx == (1, 0)
    assert xy_(1, 0).latlon == (0, 1)
    assert xy_(10, 11).x == 10
    assert xy_(10, 11).y == 11
    assert xy_(10, 11).lon == 10
    assert xy_(10, 11).lat == 11
    assert yx_(xy_(0, 1)) == xy_(yx_(1, 0))

    assert yx_(10, 20).shape == (10, 20)
    assert xy_(10, 20).wh == (10, 20)

    # equality only defined with instances of XY
    assert xy_(10, 20) != [10, 20]

    # hash
    assert len(set([xy_(0, 1), yx_(1, 0)])) == 1

    assert repr(xy_(0, 1)) == "XY(x=0, y=1)"
    assert str(xy_(0, 1)) == "XY(x=0, y=1)"

    assert ixy_(0, 3) == xy_(0, 3)
    assert iyx_(0, 3) == yx_(0, 3)

    tt = ixy_(0, 1)
    assert ixy_(tt.xy) == tt
    assert iyx_(tt.yx) == tt
    assert ixy_(tt) is tt
    assert iyx_(tt) is tt
    assert ixy_(xy_(tt.xy)) == tt
    assert iyx_(yx_(tt.yx)) == tt

    assert repr(ixy_(0, 1)) == "Index2d(x=0, y=1)"
    assert str(ixy_(0, 1)) == "Index2d(x=0, y=1)"

    assert resyx_(-10, 10) == res_(10)
    assert res_(10).x == 10
    assert res_(10).y == -10
    assert resxy_(10, -10) == resyx_(-10, 10)
    assert str(res_(10)) == "Resolution(x=10, y=-10)"
    assert repr(res_(10)) == "Resolution(x=10, y=-10)"


def test_shape2d():
    assert wh_(3, 2) == shape_((2, 3))
    assert wh_(3, 2) == (2, 3)
    wh34 = wh_(3, 4)
    assert shape_(wh34) is wh34
    assert shape_(wh34.yx) == wh34
    assert wh34 == (4, 3)
    assert shape_([4, 5]) == (4, 5)
    assert (4, 5) == shape_([4, 5])
    assert wh_(2, 3) == wh_(4, 6).shrink2()
    assert wh_(2, 3) == wh_(4, 7).shrink2()

    # should unpack like a tuple
    ny, nx = wh34
    assert (nx, ny) == wh34.xy

    # should concat like a tuple
    assert wh34 + (1,) == (4, 3, 1)
    assert ("bob",) + wh34 == ("bob", 4, 3)

    # should support len()
    assert len(wh34) == 2

    # should support indexing and slicing
    assert wh34[0] == 4
    assert wh34[::-1] == (3, 4)

    # should be a Sequence
    assert isinstance(wh34, abc.Sequence)

    # test to string
    assert str(wh34) == "Shape2d(x=3, y=4)"
    assert repr(wh34) == "Shape2d(x=3, y=4)"


def test_bad_inputs():
    # shape is valid of ints only
    with pytest.raises(ValueError):
        _ = xy_(3.1, 2).shape
    # wh is valid of ints only
    with pytest.raises(ValueError):
        _ = xy_(3.1, 2).wh

    # not enough arguments
    for op in (xy_, yx_, ixy_, iyx_):
        with pytest.raises(ValueError):
            _ = op(0)

    # making resolution from tuple should raise an error
    with pytest.raises(ValueError):
        _ = res_((-1, 2))

    with pytest.raises(ValueError):
        _ = shape_(3)


def test_map():
    assert xy_(1, 2).map(lambda x: x + 1) == xy_(2, 3)
    assert xy_(1, 2).map(lambda x: [x]) == xy_([1], [2])


def test_geom_interop():
    assert xy_(1.0, 2.0) == xy_(point(1.0, 2.0, "epsg:4326"))


def test_func_to_map() -> None:
    def aa(idx: int) -> Tuple[int, int]:
        return (idx, idx + 1)

    def first(idx: Tuple[int, ...]) -> int:
        return idx[0]

    AA = func2map(aa)
    assert len(AA) == 0
    assert list(AA) == []
    assert AA[10] == aa(10)

    FF = func2map(first)
    assert FF[3, 2] == 3
    assert FF[100, 3, 4] == 100



from typing import Tuple

import numpy as np
import pytest

from odc.geo._blocks import BlockAssembler
from odc.geo.roi import RoiTiles, roi_normalise, roi_shape, roi_tiles


@pytest.mark.parametrize(
    "tiles, idx",
    [
        (
            roi_tiles((101, 203), (11, 13)),
            [(0, 0), (1, 1), (2, 3)],
        )
    ],
)
@pytest.mark.parametrize("dtype", ["int8", "uint16", "float32"])
def test_block_assembler(tiles: RoiTiles, idx, dtype):
    dtype = np.dtype(dtype)
    Z = np.zeros(tiles.base.shape, dtype)
    for v, _i in enumerate(idx, start=1):
        Z[tiles[_i]] = v
    Zf = Z.astype(np.float32).copy()
    Zf[Z == 0] = np.nan

    blocks = {i: Z[tiles[i]].copy() for i in idx}
    ba = BlockAssembler(blocks, tiles.chunks)
    assert ba.shape == Z.shape
    assert ba.ndim == Z.ndim
    assert ba.dtype == dtype

    np.testing.assert_equal(Z, ba.extract(0))

    # for ints default fill value is 0
    if np.issubdtype(dtype, np.integer):
        np.testing.assert_equal(Z, ba.extract())
        np.testing.assert_equal(Z, ba[:, :])
        np.testing.assert_equal(Z[:10, :3], ba[:10, :3])
        np.testing.assert_equal(Z[1:-1, 2:-3], ba[1:-1, 2:-3])
        np.testing.assert_array_equal(ba.extract(dtype="float32"), Zf)

    # for floats default fill value is nan
    if np.issubdtype(dtype, np.floating):
        np.testing.assert_array_equal(Zf, ba.extract())
        np.testing.assert_array_equal(Zf, ba[:, :])
        np.testing.assert_equal(Zf[:10, 3:], ba[:10, 3:])

    # test empty
    ba = BlockAssembler({}, tiles.chunks)
    assert ba.shape == tiles.base.shape
    assert ba.ndim == 2
    assert ba.dtype == "float32"

    # axis in the wrong position
    with pytest.raises(ValueError):
        _ = BlockAssembler(blocks, chunks=tiles.chunks, axis=1)

    ii = idx[-1]

    # ndim mismatch
    bb = dict(blocks.items())
    bb[ii] = bb[ii][..., np.newaxis]
    with pytest.raises(ValueError):
        _ = BlockAssembler(bb, chunks=tiles.chunks)

    # chunk size mismatch
    bb = dict(blocks.items())
    bb[ii] = bb[ii][:-2]
    with pytest.raises(ValueError):
        _ = BlockAssembler(bb, chunks=tiles.chunks)


@pytest.mark.parametrize(
    "bshape, axis",
    [
        ((10, 13), 0),
        ((1, 10, 13), 1),
        ((10, 13, 3), 0),
    ],
)
@pytest.mark.parametrize(
    "yx_crop",
    [
        np.s_[3:4, 5:9],
        np.s_[:4, -3:],
    ],
)
def test_block_planes(bshape: Tuple[int, ...], axis: int, yx_crop):
    ny, nx = bshape[axis : axis + 2]
    NY = ny * 10 + 1
    NX = nx * 9 + min(nx - 1, 2)
    rtiles = roi_tiles((NY, NX), (ny, nx))
    chunks = rtiles.chunks
    ba = BlockAssembler({(0, 0): np.zeros(bshape)}, chunks, axis=axis)
    assert ba.ndim == len(bshape)

    for _roi in ba.planes_yx():
        assert len(_roi) == len(bshape)
        assert ba[_roi].shape == (NY, NX)

        if axis == 1 and len(bshape) == 3:
            assert isinstance(_roi[0], int)
            np.testing.assert_array_equal(ba[_roi[0]], ba[_roi])
        if axis == 0 and len(bshape) == 3:
            assert isinstance(_roi[2], int)
            np.testing.assert_array_equal(ba[:, :, _roi[2]], ba[_roi])

    crop_shape = roi_shape(roi_normalise(yx_crop, (NY, NX)))
    for _roi in ba.planes_yx(yx_crop):
        assert len(_roi) == len(bshape)
        assert ba[_roi].shape == crop_shape
        if axis == 1:
            assert isinstance(_roi[0], int)
            if len(bshape) == 3:
                np.testing.assert_array_equal(ba[_roi[0]][yx_crop], ba[_roi])

    with pytest.raises(IndexError):
        _ = ba[tuple(slice(None) for _ in range(len(bshape) + 1))]



"""Tests for the Azure AzMultiPartUpload class."""

import base64
from unittest.mock import MagicMock, patch

import pytest

pytest.importorskip("azure.storage.blob")
from odc.geo.cog._az import AzMultiPartUpload  # noqa: E402


@pytest.fixture
def azure_mpu():
    """Fixture for initializing AzMultiPartUpload."""
    account_url = "https://account_name.blob.core.windows.net"
    return AzMultiPartUpload(account_url, "container", "some.blob", None)


def test_mpu_init(azure_mpu):
    """Basic test for AzMultiPartUpload initialization."""
    assert azure_mpu.account_url == "https://account_name.blob.core.windows.net"
    assert azure_mpu.container == "container"
    assert azure_mpu.blob == "some.blob"
    assert azure_mpu.credential is None


@patch("odc.geo.cog._az.BlobServiceClient")
def test_azure_multipart_upload(mock_blob_service_client):
    """Test the full Azure AzMultiPartUpload functionality."""
    # Mock Azure Blob SDK client structure
    mock_blob_client = MagicMock()
    mock_container_client = MagicMock()
    mock_blob_service_client.return_value.get_container_client.return_value = (
        mock_container_client
    )
    mock_container_client.get_blob_client.return_value = mock_blob_client

    # Simulate return values for Azure Blob SDK methods
    mock_blob_client.get_blob_properties.return_value.etag = "mock-etag"

    # Test parameters
    account_url = "https://mockaccount.blob.core.windows.net"
    container = "mock-container"
    blob = "mock-blob"
    credential = "mock-sas-token"

    # Create an instance of AzMultiPartUpload and call its methods
    azure_upload = AzMultiPartUpload(account_url, container, blob, credential)
    upload_id = azure_upload.initiate()
    part1 = azure_upload.write_part(1, b"first chunk of data")
    part2 = azure_upload.write_part(2, b"second chunk of data")
    etag = azure_upload.finalise([part1, part2])

    # Define block IDs
    block_id1 = base64.b64encode(b"block-1").decode("utf-8")
    block_id2 = base64.b64encode(b"block-2").decode("utf-8")

    # Verify the results
    assert upload_id == "azure-block-upload"
    assert etag == "mock-etag"

    # Verify BlobServiceClient instantiation
    mock_blob_service_client.assert_called_once_with(
        account_url=account_url, credential=credential
    )

    # Verify stage_block calls
    mock_blob_client.stage_block.assert_any_call(
        block_id=block_id1, data=b"first chunk of data"
    )
    mock_blob_client.stage_block.assert_any_call(
        block_id=block_id2, data=b"second chunk of data"
    )

    # Verify commit_block_list was called correctly
    block_list = mock_blob_client.commit_block_list.call_args[0][0]
    assert len(block_list) == 2
    assert block_list[0].id == block_id1
    assert block_list[1].id == block_id2
    mock_blob_client.commit_block_list.assert_called_once()



import math
from typing import Tuple

import numpy as np
import numpy.testing as npt
import pytest
from affine import Affine

from odc.geo import XY, resxy_, xy_
from odc.geo.math import (
    Bin1D,
    Poly2d,
    affine_from_axis,
    align_down,
    align_down_pow2,
    align_up,
    align_up_pow2,
    apply_affine,
    data_resolution_and_offset,
    is_almost_int,
    is_nodata_empty,
    maybe_int,
    maybe_zero,
    quasi_random_r2,
    resolution_from_affine,
    resolve_fill_value,
    resolve_nodata,
    snap_affine,
    snap_grid,
    snap_scale,
    split_float,
    split_translation,
)
from odc.geo.testutils import mkA

NaN = float("nan")


def test_math_ops():
    assert align_up(32, 16) == 32
    assert align_up(31, 16) == 32
    assert align_up(17, 16) == 32
    assert align_up(9, 3) == 9
    assert align_up(8, 3) == 9

    assert align_down(32, 16) == 32
    assert align_down(31, 16) == 16
    assert align_down(17, 16) == 16
    assert align_down(9, 3) == 9
    assert align_down(8, 3) == 6

    assert maybe_zero(0, 1e-5) == 0
    assert maybe_zero(1e-8, 1e-5) == 0
    assert maybe_zero(-1e-8, 1e-5) == 0
    assert maybe_zero(0.1, 1e-2) == 0.1

    assert maybe_int(37, 1e-6) == 37
    assert maybe_int(37 + 1e-8, 1e-6) == 37
    assert maybe_int(37 - 1e-8, 1e-6) == 37
    assert maybe_int(3.4, 1e-6) == 3.4
    assert str(maybe_int(float("nan"), 1e-6)) == str(float("nan"))
    assert str(maybe_int(float("inf"), 1e-6)) == str(float("inf"))
    assert str(maybe_int(float("-inf"), 1e-6)) == str(float("-inf"))

    assert is_almost_int(129, 1e-6)
    assert is_almost_int(129 + 1e-8, 1e-6)
    assert is_almost_int(-129, 1e-6)
    assert is_almost_int(-129 + 1e-8, 1e-6)
    assert is_almost_int(0.3, 1e-6) is False
    assert is_almost_int(float("nan"), 1e-6) is False
    assert is_almost_int(float("inf"), 1e-6) is False
    assert is_almost_int(float("-inf"), 1e-6) is False


def test_snap_scale():
    assert snap_scale(0) == 0

    assert snap_scale(1 + 1e-6, 1e-2) == 1
    assert snap_scale(1 - 1e-6, 1e-2) == 1

    assert snap_scale(0.5 + 1e-8, 1e-3) == 0.5
    assert snap_scale(0.5 - 1e-8, 1e-3) == 0.5
    assert snap_scale(0.5, 1e-3) == 0.5

    assert snap_scale(0.6, 1e-8) == 0.6

    assert snap_scale(3.478, 1e-6) == 3.478


def test_data_res():
    xx = np.asarray([1, 2, 3, 4])
    assert data_resolution_and_offset(xx) == (1, 0.5)
    assert data_resolution_and_offset(xx[1:]) == (1, xx[1] - 1 / 2)
    assert data_resolution_and_offset(xx[:1], 1) == (1, 0.5)

    with pytest.raises(ValueError):
        data_resolution_and_offset(xx[:1])

    with pytest.raises(ValueError):
        data_resolution_and_offset(xx[:0])


def test_affine_from_axis():
    res = 10
    x0, y0 = 111, 212
    xx = np.arange(11) * res + x0 + res / 2
    yy = np.arange(13) * res + y0 + res / 2

    assert affine_from_axis(xx, yy) == Affine(res, 0, x0, 0, res, y0)

    assert affine_from_axis(xx, yy[::-1]) == Affine(
        res, 0, x0, 0, -res, yy[-1] + res / 2
    )

    # equivalent to y:-res, x:+res
    assert affine_from_axis(xx[:1], yy[:1], res) == Affine(
        res, 0, x0, 0, -res, y0 + res
    )
    assert affine_from_axis(xx[:1], yy[:1], resxy_(res, res)) == Affine(
        res, 0, x0, 0, res, y0
    )


def _check_bin(b: Bin1D, idx, tol=1e-8, nsteps=10):
    if isinstance(idx, int):
        idx = [idx]

    for _idx in idx:
        _in, _out = b[_idx]
        for x in np.linspace(_in + tol, _out - tol, nsteps):
            assert b.bin(x) == _idx


def test_bin1d_basic():
    b = Bin1D(sz=10, origin=20)
    assert b[0] == (20, 30)
    assert b[1] == (30, 40)
    assert b[-1] == (10, 20)
    assert b.bin(20) == 0
    assert b.bin(10) == -1
    assert b.bin(20 - 0.1) == -1

    for idx in [-3, -1, 0, 1, 2, 11, 23]:
        assert Bin1D.from_sample_bin(idx, b[idx], b.direction) == b

    _check_bin(b, [-3, 5])

    b = Bin1D(sz=10, origin=20, direction=-1)
    assert b[0] == (20, 30)
    assert b[-1] == (30, 40)
    assert b[1] == (10, 20)
    assert b.bin(20) == 0
    assert b.bin(10) == 1
    assert b.bin(20 - 0.1) == 1
    _check_bin(b, [-3, 5])

    assert Bin1D(10) == Bin1D(10, 0)
    assert Bin1D(10) == Bin1D(10, 0, 1)
    assert Bin1D(11) != Bin1D(10, 0, 1)
    assert Bin1D(10, 3) != Bin1D(10, 0, 1)
    assert Bin1D(10, 0, -1) != Bin1D(10, 0, 1)

    for idx in [-3, -1, 0, 1, 2, 11, 23]:
        assert Bin1D.from_sample_bin(idx, b[idx], b.direction) == b

    assert Bin1D(10) != ["something"]


def test_bin1d():
    _ii = [-3, -1, 0, 1, 2, 7]
    _check_bin(Bin1D(13.3, 23.5), _ii)
    _check_bin(Bin1D(13.3, 23.5, -1), _ii)


def test_apply_affine():
    A = mkA(rot=10, scale=(3, 1.3), translation=(-100, +2.3))
    xx, yy = np.meshgrid(np.arange(13), np.arange(11))

    xx_, yy_ = apply_affine(A, xx, yy)

    assert xx_.shape == xx.shape
    assert yy_.shape == xx.shape

    xy_expect = [A * (x, y) for x, y in zip(xx.ravel(), yy.ravel())]
    xy_got = list(zip(xx_.ravel(), yy_.ravel()))

    np.testing.assert_array_almost_equal(xy_expect, xy_got)


def test_split_translation():
    def verify(
        a: Tuple[XY[float], XY[float]],
        b: Tuple[XY[float], XY[float]],
    ):
        assert a[0].xy == pytest.approx(b[0].xy)
        assert a[1].xy == pytest.approx(b[1].xy)

    def tt(
        tx: float, ty: float, e_whole: Tuple[float, float], e_part: Tuple[float, float]
    ):
        expect = xy_(e_whole), xy_(e_part)
        rr = split_translation(xy_(tx, ty))
        verify(rr, expect)

    # fmt: off
    assert split_translation(xy_( 1,  2)) == (xy_( 1,  2), xy_(0, 0))
    assert split_translation(xy_(-1, -2)) == (xy_(-1, -2), xy_(0, 0))
    tt( 1.3, 2.5 , ( 1, 2), ( 0.3,  0.5 ))
    tt( 1.1, 2.6 , ( 1, 3), ( 0.1, -0.4 ))
    tt(-1.1, 2.8 , (-1, 3), (-0.1, -0.2 ))
    tt(-1.9, 2.05, (-2, 2), (+0.1,  0.05))
    tt(-1.5, 2.45, (-1, 2), (-0.5,  0.45))
    # fmt: on


def test_snap_affine():
    A = mkA(rot=0.1)
    assert snap_affine(A) is A

    assert snap_affine(mkA(translation=(10, 20))) == mkA(translation=(10, 20))

    assert snap_affine(mkA(translation=(10.1, 20.1)), ttol=0.2) == mkA(
        translation=(10, 20)
    )

    assert snap_affine(
        mkA(scale=(3.3, 4.2), translation=(10.1, 20.1)), ttol=0.2
    ) == mkA(scale=(3.3, 4.2), translation=(10, 20))

    assert snap_affine(
        mkA(scale=(3 + 1e-6, 4 - 1e-6), translation=(10.1, 20.1)), ttol=0.2, stol=1e-3
    ) == mkA(scale=(3, 4), translation=(10, 20))

    assert snap_affine(
        mkA(scale=(1 / 2 + 1e-8, 1 / 3 - 1e-8), translation=(10.1, 20.1)),
        ttol=0.2,
        stol=1e-3,
    ) == mkA(scale=(1 / 2, 1 / 3), translation=(10, 20))


@pytest.mark.parametrize(
    "left, right, res, off, expect",
    [
        (20, 30, 10, 0, (20, 1)),
        (20, 30.5, 10, 0, (20, 2)),
        (20, 31.5, 10, 0, (20, 2)),
        (20, 30, 10, 3, (13, 2)),
        (20, 30, -10, 0, (30, 1)),
        (19.5, 30, -10, 0, (30, 2)),
        (18.5, 30, -10, 0, (30, 2)),
        (20, 30, -10, 3, (33, 2)),
    ],
)
def test_snap_grid(left, right, res, off, expect):
    assert snap_grid(left, right, res, off / abs(res)) == expect


def test_snap_grid_tol():
    assert snap_grid(0.95, 10.12, 1, 0, 0.1) == (1.0, 10)


def test_res_affine():
    assert resolution_from_affine(mkA(scale=(2, 3))).xy == (2, 3)
    assert resolution_from_affine(mkA(rot=10, scale=(2, 3))).xy == pytest.approx((2, 3))
    assert resolution_from_affine(mkA(rot=-45, scale=(10, -10))).xy == pytest.approx(
        (10, -10)
    )


@pytest.mark.parametrize(
    "ab, expect_a, expect_b",
    [
        (1.0, 1, 0),
        (1.3, 1, 0.3),
        (0.99, 1, -0.01),
        (0.6, 1, -0.4),
        (100.6, 101, -0.4),
        (2000.3, 2000, 0.3),
        (float("nan"), float("nan"), 0),
        (float("inf"), float("inf"), 0),
    ],
)
def test_split_float(ab, expect_a, expect_b):
    a, b = split_float(ab)

    assert b == pytest.approx(expect_b)
    if np.isnan(expect_a):
        assert np.isnan(a)
        return

    assert a == pytest.approx(expect_a)
    assert (a + b) == pytest.approx(ab)


@pytest.mark.parametrize(
    "n",
    [3, 1, 10, 100],
)
@pytest.mark.parametrize("ny", [30, 101])
@pytest.mark.parametrize("nx", [20, 104])
@pytest.mark.parametrize("offset", [0, 1, 103])
def test_quasi_random_r2(n, ny, nx, offset):
    xx = quasi_random_r2(n)
    assert xx.shape == (n, 2)
    assert xx.min() >= 0
    assert xx.max() < 1

    np.testing.assert_array_equal(xx, quasi_random_r2(n))

    yy = quasi_random_r2(n, offset=offset)
    assert yy.shape == (n, 2)
    assert yy.min() >= 0
    assert yy.max() < 1
    np.testing.assert_array_equal(yy, quasi_random_r2(n, offset=offset))
    if offset == 0:
        np.testing.assert_array_equal(xx, yy)
    else:
        assert not (xx == yy).all()

    xx = quasi_random_r2(n, shape=(ny, nx))
    assert xx.min() >= 0
    assert xx[:, 0].max() < nx
    assert xx[:, 1].max() < ny


@pytest.mark.parametrize(
    "pts",
    [
        quasi_random_r2(20),
        quasi_random_r2(4, offset=10),
        quasi_random_r2(8, offset=10),
        quasi_random_r2(3, offset=20),
    ],
)
def test_poly2d(pts):
    N = pts.shape[0]
    x, y = pts.T

    p = Poly2d.fit(pts, pts)

    assert p(pts).shape == pts.shape
    assert p(x, y).shape == (2, N)

    np.testing.assert_array_almost_equal(pts, p(pts))

    assert p.grid2d(np.arange(3), np.arange(4)).shape == (2, 3, 4)

    # With input transform check
    A = Affine.scale(0.93) * Affine.translation(10, -13.2)
    p_ = p.with_input_transform(A)
    x_, y_ = apply_affine(~A, x, y)
    xx, yy = p_(x_, y_)
    np.testing.assert_array_almost_equal(x, xx)
    np.testing.assert_array_almost_equal(y, yy)

    # With input transform check, rotation
    A = Affine.rotation(-133)
    p_ = p.with_input_transform(A)
    x_, y_ = apply_affine(~A, x, y)
    xx, yy = p_(x_, y_)
    np.testing.assert_array_almost_equal(x, xx)
    np.testing.assert_array_almost_equal(y, yy)


def test_poly2d_not_enough_points():
    pts = quasi_random_r2(2)

    with pytest.raises(ValueError):
        _ = Poly2d.fit(pts[:2], pts[:2])


@pytest.mark.parametrize(
    "x",
    [
        *[2**n for n in range(20)],
        *[2**n - 1 for n in range(20)],
        *[2**n + 1 for n in range(20)],
    ],
)
def test_align_up_pow2(x: int):
    y = align_up_pow2(x)
    assert isinstance(y, int)
    assert y >= x
    assert 2 ** int(math.log2(y)) == y


@pytest.mark.parametrize(
    "x",
    [
        *[2**n for n in range(20)],
        *[2**n - 1 for n in range(1, 20)],
        *[2**n + 1 for n in range(20)],
    ],
)
def test_align_down_pow2(x: int):
    y = align_down_pow2(x)
    assert isinstance(y, int)
    assert y <= x
    assert 2 ** int(math.log2(y)) == y


@pytest.mark.parametrize(
    "nodata, dtype, expect",
    [
        (None, None, None),
        (None, "float32", None),
        ("auto", None, None),
        ("auto", "float32", NaN),
        ("auto", np.dtype("float64"), NaN),
        ("auto", "uint32", None),
        ("auto", "int16", None),
        ("auto", "uint8", None),
        ("auto", "bool", None),
        ("nan", "float32", NaN),
        (NaN, "float32", NaN),
        (np.nan, "float32", NaN),
        (33, "int8", 33),
        (-99.99, "float64", -99.99),
    ],
)
def test_resolve_nodata(nodata, dtype, expect):
    npt.assert_equal(resolve_nodata(nodata, dtype), expect)
    assert resolve_nodata("auto", dtype, 13) == 13


@pytest.mark.parametrize(
    "dst_nodata, src_nodata, dtype, expect",
    [
        (None, None, "uint16", 0),
        (None, 3, "uint16", 3),
        (4, 3, "uint16", 4),
        (10, 0, "uint16", 10),
        (NaN, 0, "float32", NaN),
        (None, 0, "float32", NaN),
        (None, None, "bool", False),
        (None, None, "uint8", 0),
    ],
)
def test_resolve_fill_value(dst_nodata, src_nodata, dtype, expect):
    npt.assert_equal(resolve_fill_value(dst_nodata, src_nodata, dtype), expect)


def test_empty():
    assert is_nodata_empty(None) is True
    assert is_nodata_empty(NaN) is True
    assert is_nodata_empty(np.nan) is True
    assert is_nodata_empty(0) is False
    assert is_nodata_empty(-1) is False
    assert is_nodata_empty(-1.0) is False
    assert is_nodata_empty(np.uint8(0)) is False



from hashlib import md5
from typing import Any, Dict, List, Tuple

import numpy as np
import pytest

from odc.geo.cog._mpu import MPUChunk, SomeData, mpu_write

FakeWriteResult = Tuple[int, SomeData, Dict[str, Any]]
# pylint: disable=unbalanced-tuple-unpacking,redefined-outer-name,import-outside-toplevel


class FakeWriter:
    """
    Fake multi-part upload writer.
    """

    def __init__(self, **limits) -> None:
        self._dst: List[FakeWriteResult] = []
        self._limits = limits

    def __call__(self, part: int, data: SomeData) -> Dict[str, Any]:
        assert 1 <= part <= 10_000
        ww = {"PartNumber": part, "ETag": f'"{etag(data)}"'}
        self._dst.append((part, data, ww))
        return ww

    def finalise(self, parts: List[Dict[str, Any]]) -> Dict[str, Any]:
        return {"Parts": parts, "Writer": self}

    @property
    def min_write_sz(self) -> int:
        return self._limits.get("min_write_sz", 5 * (1 << 20))

    @property
    def max_write_sz(self) -> int:
        return self._limits.get("max_write_sz", 5 * (1 << 30))

    @property
    def min_part(self) -> int:
        return self._limits.get("min_part", 1)

    @property
    def max_part(self) -> int:
        return self._limits.get("max_part", 10_000)

    @property
    def raw_parts(self) -> List[FakeWriteResult]:
        return self._dst

    @property
    def data(self) -> bytes:
        return b"".join(data for _, data, _ in sorted(self._dst, key=lambda x: x[0]))

    @property
    def parts(self) -> List[Dict[str, Any]]:
        return [part for _, _, part in sorted(self._dst, key=lambda x: x[0])]

    def reset(self) -> "FakeWriter":
        self._dst = []
        return self


def etag(data):
    return f'"{md5(data).hexdigest()}"'


def _mb(x: float) -> int:
    return int(x * (1 << 20))


def _mk_fake_data(sz: int) -> bytes:
    return np.random.bytes(sz)


def _split(data: bytes, offsets=List[int]) -> Tuple[bytes, ...]:
    parts = []

    for sz in offsets:
        parts.append(data[:sz])
        data = data[sz:]
    if data:
        parts.append(data)

    return tuple(parts)


@pytest.fixture(scope="function")
def write(request):
    limits = getattr(request, "param", {})
    yield FakeWriter(**limits)


def test_s3_mpu_merge_small(write: FakeWriter) -> None:
    # Test situation where parts get joined and get written eventually in one chunk
    data = _mk_fake_data(100)
    da, db, dc = _split(data, [10, 20])

    a = MPUChunk(1, 10)
    b = MPUChunk(12, 2)
    c = MPUChunk(14, 3, is_final=True)
    assert a.is_final is False
    assert b.is_final is False
    assert c.is_final is True
    assert a.started_write is False

    assert a.maybe_write(write, _mb(5)) == 0
    assert a.started_write is False

    b.append(db, "b1")
    c.append(dc, "c1")
    a.append(da, "a1")

    # b + c
    bc = MPUChunk.merge(b, c, write)
    assert bc.is_final is True
    assert bc.nextPartId == b.nextPartId
    assert bc.write_credits == (b.write_credits + c.write_credits)
    assert bc.started_write is False
    assert bc.left_data == b""
    assert bc.data == (db + dc)

    # a + (b + c)
    abc = MPUChunk.merge(a, bc, write)

    assert abc.is_final is True
    assert abc.started_write is False
    assert abc.data == data
    assert abc.left_data == b""
    assert abc.nextPartId == a.nextPartId
    assert abc.write_credits == (a.write_credits + b.write_credits + c.write_credits)
    assert len(abc.observed) == 3
    assert abc.observed == [(len(da), "a1"), (len(db), "b1"), (len(dc), "c1")]

    assert len(write.parts) == 0

    assert abc.flush_rhs(write) == len(data)
    assert len(abc.parts) == 1
    assert len(write.parts) == 1
    pid, _data, part = write.raw_parts[0]
    assert _data == data
    assert pid == a.nextPartId
    assert part["PartNumber"] == pid
    assert abc.parts[0] == part


def test_mpu_multi_writes(write: FakeWriter) -> None:
    data = _mk_fake_data(_mb(20))
    da1, db1, db2, dc1 = _split(
        data,
        [
            _mb(5.2),
            _mb(7),
            _mb(6),
        ],
    )
    a = MPUChunk(1, 10)
    b = MPUChunk(10, 2)
    c = MPUChunk(12, 3, is_final=True)
    assert (a.is_final, b.is_final, c.is_final) == (False, False, True)

    b.append(db1, "b1")
    # have enough data to write, but not enough to have left-over
    assert b.maybe_write(write, _mb(6)) == 0
    b.append(db2, "b2")
    assert b.observed == [(len(db1), "b1"), (len(db2), "b2")]
    # have enough data to write and have left-over still
    #  (7 + 6) - 6 > 5
    assert b.maybe_write(write, _mb(6)) > 0
    assert b.started_write is True
    assert len(b.data) == _mb(5)
    assert b.write_credits == 1
    assert b.nextPartId == 11

    a.append(da1, "a1")
    assert a.maybe_write(write, _mb(6)) == 0

    c.append(dc1, "c1")
    assert c.maybe_write(write, _mb(6)) == 0

    # Should flush a
    assert len(write.parts) == 1
    ab = MPUChunk.merge(a, b, write)
    assert len(write.parts) == 2
    assert len(ab.parts) == 2
    assert ab.started_write is True
    assert ab.is_final is False
    assert ab.data == (db1 + db2)[-_mb(5) :]
    assert ab.left_data == b""

    abc = MPUChunk.merge(ab, c, write)
    assert len(abc.parts) == 2
    assert abc.is_final is True
    assert abc.observed == [
        (len(da1), "a1"),
        (len(db1), "b1"),
        (len(db2), "b2"),
        (len(dc1), "c1"),
    ]

    assert abc.flush_rhs(write, None) > 0
    assert len(abc.parts) == 3
    assert write.parts == abc.parts
    assert write.data == data


def test_mpu_left_data(write: FakeWriter) -> None:
    data = _mk_fake_data(_mb(3 + 2 + (6 + 5.2)))
    da1, db1, dc1, dc2 = _split(
        data,
        [
            _mb(3),
            _mb(2),
            _mb(6),
        ],
    )
    a = MPUChunk(1, 100)
    b = MPUChunk(100, 100)
    c = MPUChunk(200, 100)

    c.append(dc1, "c1")
    c.append(dc2, "c2")
    assert c.maybe_write(write, _mb(6)) > 0
    assert c.started_write is True
    assert len(c.parts) == 1
    assert c.left_data == b""

    b.append(db1, "b1")
    a.append(da1, "a1")

    bc = MPUChunk.merge(b, c, write)
    assert bc.started_write is True
    assert len(bc.left_data) > 0
    assert bc.nextPartId == 201
    assert bc.write_credits == 99

    # Expect (a.data + bc.left_data) to be written to PartId=1
    abc = MPUChunk.merge(a, bc, write)
    assert len(abc.parts) == 2
    assert abc.parts[0]["PartNumber"] == 1
    assert abc.nextPartId == 201
    assert abc.write_credits == 99
    assert abc.is_final is False

    assert abc.observed == [
        (len(da1), "a1"),
        (len(db1), "b1"),
        (len(dc1), "c1"),
        (len(dc2), "c2"),
    ]

    assert abc.flush_rhs(write) > 0
    assert abc.nextPartId == 202
    assert abc.write_credits == 98
    assert len(abc.parts) == 3
    assert write.data == data
    assert write.parts == abc.parts


def test_mpu_misc(write: FakeWriter) -> None:
    a = MPUChunk(1, 10)
    b = MPUChunk(10, 1)

    data = _mk_fake_data(_mb(3 + (6 + 7)))
    da1, db1, db2 = _split(
        data,
        [_mb(3), _mb(6), _mb(7)],
    )
    b.append(db1, "b1")
    b.append(db2, "b2")

    # not enough credits to write
    assert b.maybe_write(write, _mb(5)) == 0

    a.append(da1, "a1")

    ab = MPUChunk.merge(a, b, write)
    assert ab.started_write is False
    assert len(write.parts) == 0
    assert ab.nextPartId == 1
    assert ab.write_credits == 11

    assert ab.flush_rhs(write) > 0
    assert len(ab.parts) == 1
    assert write.data == data
    assert write.parts == ab.parts

    assert " final" in repr(
        MPUChunk(
            1,
            1,
            bytearray(),
            parts=[{}],
            observed=[(0, None)],
            is_final=True,
        )
    )


def test_lhs_keep(write: FakeWriter) -> None:
    def mpu(lhs_keep: int, extra_bytes: int, is_final: bool) -> Tuple[MPUChunk, bytes]:
        a = MPUChunk(10, 100, lhs_keep=lhs_keep, is_final=is_final)
        a.append(data := _mk_fake_data(lhs_keep + extra_bytes))
        return a, data

    lhs_keep = write.min_write_sz

    a, data = mpu(lhs_keep, write.min_write_sz // 2, is_final=False)
    assert a.left_data == b""
    assert a.data == data
    assert not a.started_write

    # should move all data .left_data
    assert a.flush_rhs(write) == 0
    assert a.data == b""
    assert a.left_data == data

    a, data = mpu(lhs_keep, write.min_write_sz // 2, is_final=True)
    assert a.left_data == b""
    assert a.data == data
    assert not a.started_write

    # should not spill: not enough due to lhs
    assert a.maybe_write(write, write.min_write_sz) == 0
    assert not a.started_write

    assert a.flush_rhs(write) == len(data) - lhs_keep
    assert a.started_write
    assert a.data == b""
    assert a.left_data == data[:lhs_keep]


@pytest.mark.parametrize("spill_sz", [10, 11, 10_000, 0])
def test_dask_parts(spill_sz: int):
    pytest.importorskip("dask")
    from dask import bag
    from dask.delayed import Delayed

    write = FakeWriter(min_write_sz=10)
    assert write.min_write_sz == 10
    assert spill_sz == 0 or spill_sz >= write.min_write_sz

    data = _mk_fake_data(103)
    parts = _split(data, [30, 50])
    parts = [(bb, idx) for idx, bb in enumerate(parts)]

    chunks = bag.from_sequence(parts, npartitions=len(parts))

    FOOTER = b"\n------------\n"
    HEADER = b"Header\n\n"

    def mk_footer(observed, expected_sz: int = 0):
        # total sum of observed elements should match data length
        assert sum(sz for sz, _ in observed) == expected_sz
        return FOOTER

    def mk_header(observed, expected_sz: int = 0):
        # total sum of observed elements should match data length
        assert sum(sz for sz, _ in observed) == expected_sz
        return HEADER

    # check single bag case with writer
    rr = mpu_write(chunks, write, spill_sz=spill_sz)
    assert isinstance(rr, Delayed)
    rr = rr.compute()
    assert isinstance(rr, dict)
    assert rr["Writer"] is write
    assert rr["Parts"] == write.parts
    assert write.data == data

    # check substreams + footer + header
    rr = mpu_write(
        [chunks] * 3,
        write=write.reset(),
        spill_sz=spill_sz,
        mk_footer=mk_footer,
        mk_header=mk_header,
        user_kw={"expected_sz": len(data) * 3},
    )
    assert isinstance(rr, Delayed)
    rr = rr.compute()
    assert isinstance(rr, dict)
    assert rr["Writer"] is write
    assert rr["Parts"] == write.parts
    assert write.data == b"".join([HEADER, data * 3, FOOTER])

    # check no writer case
    rr = mpu_write(chunks)
    assert isinstance(rr, Delayed)
    rr = rr.compute()
    assert isinstance(rr, MPUChunk)
    assert rr.is_final
    assert rr.lhs_keep == 0
    assert rr.left_data == bytearray()
    assert rr.data == data
    assert rr.started_write is False
    assert rr.parts == []



"""
Interop with other geometry libraries.
"""

import re
from typing import Any, List, Tuple, Union

from .crs import CRS, MaybeCRS, Optional, norm_crs
from .gcp import GCPGeoBox
from .geobox import GeoBox
from .geom import Geometry, point
from .types import XY, xy_


def from_geopandas(series) -> List[Geometry]:
    """
    Convert Geopandas data into list of :py:class:`~odc.geo.geom.Geometry`.
    """
    crs = getattr(series, "crs", None)
    gg = getattr(series, "geometry", None)

    if crs is None or gg is None:
        return []

    return [Geometry(g, crs) for g in gg]


def extract_gcps_raw(src) -> Tuple[List[XY[float]], List[XY[float]], CRS]:
    """
    Extract Ground Control points from :py:class:`rasterio.DatasetReader`.

    :returns: ``[pixel coors], [world coords], CRS``
    """
    pts, gcp_crs = src.gcps
    if len(pts) == 0:
        raise ValueError(f"Could not access GCP on {src}")

    pix_pts = [xy_(pt.col, pt.row) for pt in pts]
    wld_pts = [xy_(pt.x, pt.y) for pt in pts]

    return pix_pts, wld_pts, norm_crs(gcp_crs)


def extract_gcps(
    src,
    output_crs: MaybeCRS = None,
) -> Tuple[List[XY[float]], List[Geometry]]:
    """
    Extract Ground Control points from :py:class:`rasterio.DatasetReader`.

    :returns: ``[pixel coords], [world coords]``
    """
    pix, wld, gcp_crs = extract_gcps_raw(src)
    output_crs = norm_crs(output_crs)

    wld_pts = (point(pt.x, pt.y, gcp_crs) for pt in wld)
    if output_crs is not None and gcp_crs != output_crs:
        wld_pts = (pt.to_crs(output_crs) for pt in wld_pts)

    return pix, list(wld_pts)


def map_crs(m: Any, /) -> Optional[CRS]:
    def _from_name(srs: str) -> Optional[CRS]:
        if match := re.match(r"epsg:?(?P<code>\d+)", srs.lower()):
            return CRS(f"epsg:{match['code']}")
        return None

    _crs = getattr(m, "crs", None)
    if isinstance(_crs, str):
        # probably folium map
        return _from_name(_crs)

    if isinstance(_crs, dict):
        # ipylealflet uses dict
        if (name := _crs.get("name", None)) is not None:
            if (crs := _from_name(name)) is not None:
                return crs
        if (proj4def := _crs.get("proj4def", None)) is not None:
            return CRS(proj4def)

    return None


def rio_geobox(rdr: Any) -> Union[GeoBox, GCPGeoBox]:
    """
    Construct GeoBox from rasterio.

    :param rdr: Opened :py:class:`rasterio.DatasetReader`
    :returns:
       :py:class:`~odc.geo.geobox.GeoBox` or :py:class:`~odc.geo.gcp.GCPGeoBox`.
    """
    pts, _ = rdr.gcps
    if len(pts) > 0:
        return GCPGeoBox.from_rio(rdr)

    return GeoBox.from_rio(rdr)



from odc.geo.data import country_geom, data_path, gbox_css, ocean_geojson, ocean_geom


def test_ocean_gjson():
    g1 = ocean_geojson()
    g2 = ocean_geojson()

    assert g1 is g2
    assert len(g1["features"]) == 2


def test_ocean_geom():
    g = ocean_geom()
    assert g.crs == "epsg:4326"

    g = ocean_geom("epsg:3857", (-180, -80, 180, 80))
    assert g.crs == "epsg:3857"


def test_country_geom():
    g = country_geom("AUS")
    assert g.crs == "epsg:4326"

    g = country_geom("AUS", "epsg:3577")
    assert g.crs == "epsg:3577"


def test_gbox_css():
    assert isinstance(gbox_css(), str)


def test_data_path():
    assert data_path().exists()
    assert data_path("--no-such-thing--").exists() is False



import numpy as np
import pytest

from odc.geo.roi import (
    RoiTiles,
    Tiles,
    VariableSizedTiles,
    _norm_slice_or_error,
    clip_tiles,
    polygon_path,
    roi_boundary,
    roi_center,
    roi_from_points,
    roi_intersect,
    roi_is_empty,
    roi_is_full,
    roi_normalise,
    roi_pad,
    roi_shape,
    roi_tiles,
    scaled_down_roi,
    scaled_down_shape,
    scaled_up_roi,
    slice_intersect3,
    w_,
)


def test_roi_tools():
    s_ = np.s_

    assert roi_shape(s_[2:4, 3:4]) == (2, 1)
    assert roi_shape(s_[:4, :7]) == (4, 7)
    assert roi_shape(s_[3, :7]) == (1, 7)

    assert roi_is_empty(s_[:4, :5]) is False
    assert roi_is_empty(s_[1:1, :10]) is True
    assert roi_is_empty(s_[7:3, :10]) is True

    assert roi_is_empty(s_[:3]) is False
    assert roi_is_empty(s_[4:4]) is True

    assert roi_is_full(s_[:3], 3) is True
    assert roi_is_full(s_[:3, 0:4], (3, 4)) is True
    assert roi_is_full(s_[:, 0:4], (33, 4)) is True
    assert roi_is_full(s_[1:3, 0:4], (3, 4)) is False
    assert roi_is_full(s_[1:3, 0:4], (2, 4)) is False
    assert roi_is_full(s_[0:4, 0:4], (3, 4)) is False
    assert roi_is_full(s_[0], 1) is True
    assert roi_is_full(s_[2], 3) is False

    roi = s_[0:8, 0:4]
    roi_ = scaled_down_roi(roi, 2)
    assert roi_shape(roi_) == (4, 2)
    assert scaled_down_roi(scaled_up_roi(roi, 3), 3) == roi

    assert scaled_down_shape(roi_shape(roi), 2) == roi_shape(scaled_down_roi(roi, 2))

    assert roi_shape(scaled_up_roi(roi, 10000, (40, 50))) == (40, 50)

    for bad_roi in [np.s_[1:], np.s_[:], np.s_[-3:]]:
        with pytest.raises(ValueError):
            _ = roi_shape(bad_roi)

    assert roi_normalise(s_[3:4], 40) == s_[3:4]
    assert roi_normalise(s_[3], 40) == s_[3:4]
    assert roi_normalise(s_[:4], (40,)) == s_[0:4]
    assert roi_normalise(s_[:], (40,)) == s_[0:40]
    assert roi_normalise(s_[:-1], (3,)) == s_[0:2]
    assert roi_normalise(-1, (3,)) == s_[2:3]
    assert roi_normalise(s_[-1:], (3,)) == s_[2:3]
    assert roi_normalise((s_[:-1],), 3) == (s_[0:2],)
    assert roi_normalise(s_[-2:-1, :], (10, 20)) == s_[8:9, 0:20]
    assert roi_normalise(s_[-2:-1, :, 3:4], (10, 20, 100)) == s_[8:9, 0:20, 3:4]
    assert roi_center(s_[0:3]) == 1.5
    assert roi_center(s_[0:2, 0:6]) == (1, 3)


def test_roi_from_points():
    roi = np.s_[0:2, 4:13]
    xy = roi_boundary(roi)

    assert xy.shape == (4, 2)
    assert roi_from_points(xy, (2, 13)) == roi

    xy = np.asarray(
        [
            [0.2, 1.9],
            [10.3, 21.2],
            [float("nan"), 11],
            [float("inf"), float("-inf")],
        ]
    )
    assert roi_from_points(xy, (100, 100)) == np.s_[1:22, 0:11]
    assert roi_from_points(xy, (5, 7)) == np.s_[1:5, 0:7]
    assert roi_from_points(xy[2:, :], (3, 3)) == np.s_[0:0, 0:0]


@pytest.mark.parametrize(
    "a,b",
    [
        np.s_[0:3, :3],
        np.s_[:4, 2:6],
        np.s_[4:13, 5:17],
        np.s_[10:13, 3:7],
        np.s_[10:13, 13:17],
        np.s_[10:13, 14:17],
    ],
)
def test_slice_intersect3(a: slice, b: slice):
    assert isinstance(a.stop, int)
    assert isinstance(b.stop, int)
    _a, _b, _ab = slice_intersect3(a, b)

    (na,) = roi_shape(a)
    (nb,) = roi_shape(b)

    assert _a.start <= _a.stop
    assert 0 <= _a.start <= na
    assert 0 <= _a.stop <= na

    assert _b.start <= _b.stop
    assert 0 <= _b.start <= nb
    assert 0 <= _b.stop <= nb

    X = np.arange(max(a.stop, b.stop))
    np.testing.assert_array_equal(X[a][_a], X[b][_b])
    np.testing.assert_array_equal(X[a][_a], X[_ab])


def test_roi_intersect():
    s_ = np.s_
    roi = s_[0:2, 4:13]

    assert roi_intersect(roi, roi) == roi
    assert roi_intersect(s_[0:3], s_[1:7]) == s_[1:3]
    assert roi_intersect(s_[0:3], (s_[1:7],)) == s_[1:3]
    assert roi_intersect((s_[0:3],), s_[1:7]) == (s_[1:3],)

    assert roi_intersect(s_[4:7, 5:6], s_[0:1, 7:8]) == s_[4:4, 6:6]


def test_roi_pad():
    s_ = np.s_
    assert roi_pad(s_[0:4], 1, 4) == s_[0:4]
    assert roi_pad(s_[0:4], 1, (4,)) == s_[0:4]
    assert roi_pad((s_[0:4],), 1, 4) == (s_[0:4],)

    assert roi_pad(s_[0:4, 1:5], 1, (4, 6)) == s_[0:4, 0:6]
    assert roi_pad(s_[2:3, 1:5], 10, (7, 9)) == s_[0:7, 0:9]
    assert roi_pad(s_[3, 0, :2], 1, (100, 100, 100)) == s_[2:5, 0:2, 0:3]


def test_norm_slice_or_error():
    s_ = np.s_
    assert _norm_slice_or_error(s_[0]) == s_[0:1]
    assert _norm_slice_or_error(s_[3]) == s_[3:4]
    assert _norm_slice_or_error(s_[:3]) == s_[0:3]
    assert _norm_slice_or_error(s_[10:100:3]) == s_[10:100:3]

    for bad in [np.s_[1:], np.s_[:-3], np.s_[-3:], np.s_[-2:10], -3]:
        with pytest.raises(ValueError):
            _ = _norm_slice_or_error(bad)


def test_window_from_slice():
    s_ = np.s_

    assert w_[None] is None
    assert w_[s_[:3, 4:5]] == ((0, 3), (4, 5))
    assert w_[s_[0:3, :5]] == ((0, 3), (0, 5))
    assert w_[list(s_[0:3, :5])] == ((0, 3), (0, 5))

    for roi in [s_[:3], s_[:3, :4, :5], 0]:
        with pytest.raises(ValueError):
            _ = w_[roi]


def test_polygon_path():
    pp = polygon_path([0, 1])
    assert pp.shape == (2, 5)
    assert set(pp.ravel()) == {0, 1}

    pp2 = polygon_path([0, 1], [0, 1])
    assert (pp2 == pp).all()

    pp = polygon_path([0, 1], [2, 3])
    assert set(pp[0].ravel()) == {0, 1}
    assert set(pp[1].ravel()) == {2, 3}


def test_tiles():
    tt = Tiles((10, 20), (3, 7))
    assert tt.tile_shape((0, 0)) == (3, 7)
    assert tt.tile_shape((3, 2)) == (1, 6)
    assert tt.shape.yx == (4, 3)
    assert tt.base.yx == (10, 20)
    assert tt.chunks == ((3, 3, 3, 1), (7, 7, 6))

    assert tt[0, 0] == np.s_[0:3, 0:7]
    assert tt[3, 2] == np.s_[9:10, 14:20]
    assert tt[-1, -1] == tt[3, 2]
    assert tt[-4, -3] == tt[0, 0]
    assert tt.tile_shape((-1, -1)) == tt.tile_shape((3, 2))
    assert tt.tile_shape((-4, -3)) == tt.tile_shape((0, 0))

    assert tt[0, 0] == tt[:1, :1]
    assert tt[:, :] == np.s_[0:10, 0:20]
    assert tt[0:2, 0] == np.s_[0:6, 0:7]
    assert tt[1:4, -1:] == np.s_[3:10, 14:20]
    assert tt[1:, -1:] == np.s_[3:10, 14:20]

    assert tt == tt
    assert tt != Tiles((1, 1), (1, 1))
    assert tt != {"": 3}

    tt_ = VariableSizedTiles(tt.chunks)
    assert tt_.shape == tt.shape
    assert tt_.tile_shape((0, 0)) == (3, 7)
    assert tt_.tile_shape((3, 2)) == (1, 6)
    assert tt_.shape.yx == (4, 3)
    assert tt_.base.yx == (10, 20)
    assert tt_.chunks == ((3, 3, 3, 1), (7, 7, 6))
    assert tt[-1, -1] == tt[3, 2]
    assert tt[-4, -3] == tt[0, 0]
    assert tt.tile_shape((-1, -1)) == tt.tile_shape((3, 2))
    assert tt.tile_shape((-4, -3)) == tt.tile_shape((0, 0))

    assert isinstance(roi_tiles(tt.shape, (1, 2)), Tiles)
    assert isinstance(roi_tiles(tt.shape, tt.shape), Tiles)
    assert isinstance(roi_tiles(tt.shape, tt.chunks), VariableSizedTiles)

    # smoke test repr/str
    assert isinstance(repr(tt), str)
    assert isinstance(repr(tt_), str)
    assert "100_123x200_321" in repr(Tiles((100_123, 200_321), (1000, 1000)))


@pytest.mark.parametrize(
    "chunks",
    [
        ((1,), (2,)),
        ((1, 10, 3), (2, 5, 7, 11, 2)),
    ],
)
def test_varsz_tiles(chunks):
    iy, ix = chunks
    tt = VariableSizedTiles(chunks)

    assert tt.shape == (len(iy), len(ix))
    assert tt.base == (sum(iy), sum(ix))
    assert tt == tt
    assert (tt == "") is False
    assert tt != [9]

    # test comparison with different chunk sizes
    _chunks = tuple(ch + (1,) for ch in chunks)
    assert tt != VariableSizedTiles(_chunks)

    # test comparison with different chunk values
    _chunks = tuple((ch[0] + 3, *ch[1:]) for ch in chunks)
    assert tt != VariableSizedTiles(_chunks)

    for idx in np.ndindex(tt.shape.yx):
        y, x = idx
        assert tt.tile_shape(idx) == (iy[y], ix[x])
        assert (
            tt[idx]
            == np.s_[
                sum(iy[:y]) : sum(iy[: y + 1]),
                sum(ix[:x]) : sum(ix[: x + 1]),
            ]
        )
        # test slice version
        assert tt[y, x] == tt[y : y + 1, x : x + 1]
        assert tt[y, x] == tt[y, x : x + 1]
        assert tt[y, x] == tt[y : y + 1, x]
        assert tt[:y, x:] == np.s_[0 : sum(iy[:y]), sum(ix[:x]) : sum(ix)]

    assert isinstance(tt.__dask_tokenize__(), tuple)


@pytest.mark.parametrize(
    "tile",
    [
        VariableSizedTiles(((10, 1, 30), (2, 4, 5, 6))),
        VariableSizedTiles(((10, 1, 30), (2, 4, 7, 13))),
        Tiles((104, 201), (11, 23)),
    ],
)
@pytest.mark.parametrize(
    "roi",
    [
        np.s_[:1, :1],
        np.s_[1:2, 2:],
        np.s_[-1:, 0],
    ],
)
def test_tiles_crop(tile, roi):
    assert isinstance(tile.crop(roi), type(tile))

    # tile.chunks[roi] == tile[roi].chunks
    _roi = roi_normalise(roi, tile.shape.yx)
    expect_chunks = tuple(ch[s.start : s.stop] for ch, s in zip(tile.chunks, _roi))
    assert tile.crop(roi).chunks == expect_chunks


@pytest.mark.parametrize(
    "tiles",
    [
        VariableSizedTiles(((10, 1, 30), (2, 4, 5, 6))),
        VariableSizedTiles(((10, 1, 30), (2, 4, 7, 13))),
        Tiles((104, 201), (11, 23)),
    ],
)
def test_clip_tiles(tiles: RoiTiles):
    ny, nx = tiles.shape.yx
    tl, tr, br, bl = (0, 0), (0, nx - 1), (ny - 1, nx - 1), (ny - 1, 0)

    tt, roi, idx = clip_tiles(tiles, [tl, tr, br, bl])
    assert tt == tiles
    assert roi == np.s_[0:ny, 0:nx]
    assert idx == [tl, tr, br, bl]


@pytest.mark.parametrize(
    "tiles",
    [
        VariableSizedTiles(((10, 1, 30), (2, 4, 5, 6))),
        VariableSizedTiles(((10, 1, 30), (2, 4, 7, 13))),
        Tiles((104, 201), (11, 23)),
    ],
)
def test_locate(tiles: RoiTiles):
    NY, NX = tiles.base.yx
    ny, nx = tiles.shape.yx

    # check all four corners
    assert tiles.locate((0, 0)) == (0, 0)
    assert tiles.locate((NY - 1, NX - 1)) == (ny - 1, nx - 1)
    assert tiles.locate((0, NX - 1)) == (0, nx - 1)
    assert tiles.locate((NY - 1, 0)) == (ny - 1, 0)

    for idx in [(-1, 0), (NY, 0), (NY, NX), (NY * 1000, 1)]:
        with pytest.raises(IndexError):
            _ = tiles.locate(idx)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
""" Geometric shapes and operations on them
"""
# import order is important here
#  _crs <-- _geom <-- _geobox <- other
# isort: skip_file

from ._version import __version__

from .types import (
    AnchorEnum,
    XY,
    Index2d,
    Shape2d,
    Resolution,
    SomeShape,
    SomeResolution,
    SomeIndex2d,
    Unset,
    xy_,
    yx_,
    res_,
    resxy_,
    resyx_,
    ixy_,
    iyx_,
    wh_,
    shape_,
)

from .crs import (
    CRS,
    CRSError,
    CRSMismatchError,
    MaybeCRS,
    SomeCRS,
)

from .geom import (
    BoundingBox,
    CoordList,
    Geometry,
)

__all__ = [
    "__version__",
    "AnchorEnum",
    "XY",
    "Index2d",
    "Shape2d",
    "Resolution",
    "SomeShape",
    "SomeResolution",
    "SomeIndex2d",
    "BoundingBox",
    "CoordList",
    "CRS",
    "CRSError",
    "CRSMismatchError",
    "Geometry",
    "MaybeCRS",
    "SomeCRS",
    "Unset",
    "xy_",
    "yx_",
    "res_",
    "resxy_",
    "resyx_",
    "ixy_",
    "iyx_",
    "wh_",
    "shape_",
]



import numpy as np
import pytest
import xarray as xr

from odc.geo._interop import is_dask_collection
from odc.geo._rgba import is_rgb
from odc.geo.testutils import daskify
from odc.geo.xr import ODCExtensionDa, ODCExtensionDs

try:
    import matplotlib
except ImportError:
    matplotlib = None


def test_colorize(ocean_raster: xr.DataArray):
    xx = ocean_raster

    assert isinstance(xx.odc, ODCExtensionDa)
    cmap = np.asarray(
        [
            [153, 153, 102, 255],
            [51, 102, 153, 255],
        ],
        dtype="uint8",
    )

    cc = xx.odc.colorize(cmap)
    assert isinstance(cc.odc, ODCExtensionDa)
    assert cc.odc.geobox == xx.odc.geobox
    assert cc.dtype == cmap.dtype
    assert cc.shape == (*xx.shape, 4)

    cc = xx.astype("int16").odc.colorize(cmap, clip=True)
    assert isinstance(cc.odc, ODCExtensionDa)
    assert cc.odc.geobox == xx.odc.geobox
    assert cc.dtype == cmap.dtype
    assert cc.shape == (*xx.shape, 4)

    _xx = daskify(xx, chunks=(16, 32))
    assert is_dask_collection(_xx) is True

    _cc = _xx.odc.colorize(cmap, clip=True)
    assert is_dask_collection(_cc) is True

    assert isinstance(_cc.odc, ODCExtensionDa)
    assert _cc.odc.geobox == _xx.odc.geobox
    assert _cc.dtype == cmap.dtype
    assert _cc.shape == (*_xx.shape, 4)

    assert bool((_cc.compute() == cc).all()) is True


@pytest.mark.skipif(matplotlib is None, reason="Needs matplotlib")
def test_colorize_matplotlib(ocean_raster: xr.DataArray):
    xx = ocean_raster
    cc = xx.odc.colorize("jet")
    assert isinstance(cc.odc, ODCExtensionDa)
    assert cc.odc.geobox == xx.odc.geobox
    assert cc.dtype == "uint8"
    assert cc.shape == (*xx.shape, 4)

    cc = xx.astype("float32").odc.colorize(matplotlib.colormaps.get_cmap("viridis"))
    assert isinstance(cc.odc, ODCExtensionDa)
    assert cc.odc.geobox == xx.odc.geobox
    assert cc.dtype == "uint8"
    assert cc.shape == (*xx.shape, 4)

    cc = xx.astype("float32").odc.colorize()
    assert isinstance(cc.odc, ODCExtensionDa)
    assert cc.odc.geobox == xx.odc.geobox
    assert cc.dtype == "uint8"
    assert cc.shape == (*xx.shape, 4)


def test_rgba(ocean_raster_ds: xr.Dataset):
    xx = ocean_raster_ds
    assert xx.red.dtype == "int16"
    assert isinstance(xx.odc, ODCExtensionDs)
    assert xx.red.nodata == -1

    cc = xx.odc.to_rgba(vmin=0, vmax=3_000)
    assert isinstance(cc.odc, ODCExtensionDa)
    assert cc.odc.geobox == xx.odc.geobox
    assert cc.dtype == "uint8"
    assert cc.shape == (*xx.red.shape, 4)
    assert is_rgb(cc) is True
    assert is_rgb(xx.red) is False

    cc = xx.astype("float32").odc.to_rgba()
    assert isinstance(cc.odc, ODCExtensionDa)
    assert cc.odc.geobox == xx.odc.geobox
    assert cc.dtype == "uint8"
    assert cc.shape == (*xx.red.shape, 4)

    _xx = xx.rename(dict(red="b1", green="b2", blue="b3"))
    for dv in _xx.data_vars.values():
        dv.attrs.pop("nodata", None)

    cc = _xx.odc.to_rgba(["b1", "b2", "b3"], vmax=3_000)  # vmin defaults to 0
    assert isinstance(cc.odc, ODCExtensionDa)
    assert cc.odc.geobox == xx.odc.geobox
    assert cc.dtype == "uint8"
    assert cc.shape == (*xx.red.shape, 4)
    assert is_rgb(cc[..., 0]) is False
    assert is_rgb(cc[..., :2]) is False
    assert is_rgb(cc[..., :3]) is True

    # can't guess band names
    with pytest.raises(ValueError):
        _ = _xx.odc.to_rgba(vmax=3_000)

    # too many red band candidates
    _xx = xx.rename({"red": "red1"})
    _xx["red2"] = xx.red
    with pytest.raises(ValueError):
        _ = _xx.odc.to_rgba(vmax=3_000)

    _xx = daskify(xx, chunks=(51, 39))
    assert is_dask_collection(_xx) is True
    _cc = _xx.odc.to_rgba(vmax=3_000)
    assert isinstance(_cc.odc, ODCExtensionDa)
    assert is_dask_collection(_cc) is True
    assert _cc.odc.geobox == xx.odc.geobox
    assert _cc.dtype == "uint8"
    assert _cc.shape == (*xx.red.shape, 4)
    assert bool((cc == _cc.compute()).all()) is True

    # missing clamp for dask inputs
    with pytest.raises(ValueError):
        _ = _xx.odc.to_rgba()

    # smoke-test compression
    assert isinstance(cc.odc.compress("jpeg", 60, transparent=(0, 0, 0)), bytes)
    assert isinstance(_cc.odc.compress("png", 9), bytes)
    assert isinstance(cc.odc.compress(as_data_url=True), str)

    assert isinstance(xx.red.astype("uint8").odc.compress(), bytes)

    with pytest.raises(ValueError):
        _ = xx.red.odc.compress("no-such-format")

    with pytest.raises(ValueError):
        # expect 2d/3d inputs
        _ = xx.red[0].odc.compress()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import math
from dataclasses import dataclass
from typing import Callable, Literal, Optional, Protocol, Sequence, Tuple, Union

import numpy as np
from affine import Affine

from .crs import SomeCRS
from .gcp import GCPGeoBox
from .geobox import GeoBox, GeoboxAnchor, GeoBoxBase, gbox_boundary
from .math import (
    affine_from_pts,
    decompose_rws,
    is_affine_st,
    is_almost_int,
    maybe_int,
    snap_affine,
    stack_xy,
    unstack_xy,
)
from .roi import (
    NormalizedROI,
    NormalizedSlice,
    roi_boundary,
    roi_center,
    roi_from_points,
    roi_is_empty,
    scaled_up_roi,
)
from .types import XY, SomeResolution, SomeShape, res_, shape_, xy_


class PointTransform(Protocol):
    """
    Invertible point transform.
    """

    def __call__(self, pts: Sequence[XY[float]]) -> Sequence[XY[float]]: ...

    @property
    def back(self) -> "PointTransform": ...

    @property
    def linear(self) -> Optional[Affine]: ...


class LinearPointTransform:
    """
    Point transform within the same projection.
    """

    def __init__(self, A: Affine, back: Optional["LinearPointTransform"] = None):
        self.A = A
        self._back = back

    @property
    def linear(self) -> Optional[Affine]:
        return self.A

    @property
    def back(self) -> "LinearPointTransform":
        if self._back is not None:
            return self._back
        back = LinearPointTransform(~self.A, self)
        self._back = back
        return back

    def __call__(self, pts: Sequence[XY[float]]) -> Sequence[XY[float]]:
        A = self.A
        return [xy_(A * pt.xy) for pt in pts]

    def __repr__(self) -> str:
        return f"LinearPointTransform(\n  {self.A!r})"


class GbxPointTransform:
    """
    Point Transform between two pixel planes.

    Maps pixel coordinates from one geo referenced image to another.

    1. Input is source pixel coordinate

    2. Compute coordinate in CRS units using pix2wld transform of the source image

    3. Project point to the CRS of the destination image

    4. Compute destination image pixel coordinate by using wld2pix mapping of the
       destination image
    """

    def __init__(
        self,
        src: GeoBoxBase,
        dst: GeoBoxBase,
        back: Optional["GbxPointTransform"] = None,
    ):
        assert src.crs is not None and dst.crs is not None
        self._src = src
        self._dst = dst
        self._back = back
        self._tr = src.crs.transformer_to_crs(dst.crs)
        self._clamps: Optional[Tuple[Tuple[float, float], Tuple[float, float]]] = None
        if src.crs.geographic:
            self._clamps = ((-180, 180), (-90, 90))

    @property
    def back(self) -> "GbxPointTransform":
        if self._back is not None:
            return self._back
        back = GbxPointTransform(self._dst, self._src, self)
        self._back = back
        return back

    @property
    def linear(self) -> Optional[Affine]:
        return None

    def __call__(self, pts: Sequence[XY[float]]) -> Sequence[XY[float]]:
        # pix_src -> X -> X' -> pix_dst
        # inv(dst.A)*to_crs(src.A*pt)

        xx, yy = np.asarray([pt.xy for pt in pts]).T
        xx, yy = self._src.pix2wld(xx, yy)

        if self._clamps is not None:
            # for global datasets in 4326 pixel edges sometimes reach just outside
            # of the valid region due to rounding errors when creating tiff files
            # those coordinates can then not be converted properly to destintation crs
            range_x, range_y = self._clamps
            xx = np.clip(xx, *range_x)
            yy = np.clip(yy, *range_y)

        xx, yy = self._dst.wld2pix(*self._tr(xx, yy))
        return [xy_(x, y) for x, y in zip(xx, yy)]

    def __repr__(self) -> str:
        return f"GbxPointTransform({self._src!r}, {self._dst!r})"


@dataclass
class ReprojectInfo:
    """
    Describes computed data loading parameters.

    For scale direction is: "scale > 1 --> shrink src to fit dst".
    """

    roi_src: NormalizedROI
    """Section of the source image to load."""

    roi_dst: NormalizedROI
    """Section of the destination image to update."""

    paste_ok: bool
    """
    When ``True`` source can be pasted into destination directly.

    * Must have same projection

    * Must have same pixel size, or same pixel size after shrinking
      source with integer scaling

    * Sub-pixel translation between the source and the destination images must be lower
      than the requested tolerance

    """

    read_shrink: int
    """
    Highest allowed integer shrink factor on the read side.

    Used to pick overview level for reading. A value of ``3`` means you can
    shrink every ``3x3`` pixel block of the source image down to a single pixel,
    and still have higher resolution than requested.
    """

    scale: float
    """Scale change as a single number. (the min of scale2)"""

    scale2: XY[float]
    """Full 2D Scale change as an XY."""

    transform: PointTransform
    """Mapping from src pixels to destination pixels."""


def get_scale_from_linear_transform(A: Affine) -> XY[float]:
    """
    Given a linear transform compute scale change.

    1. Y = A*X + t
    2. Extract scale components of A

    Returns (sx, sy), where sx > 0, sy > 0
    """
    _, _, S = decompose_rws(A)
    return xy_(abs(S.a), abs(S.e))


def get_scale_at_point(
    pt: XY[float], tr: PointTransform, r: Optional[float] = None
) -> XY[float]:
    """
    Given an arbitrary locally linear transform estimate scale change around a point.

    1. Approximate ``Y = tr(X)`` as ``Y = A*X + t`` in the neighbourhood of pt, for ``X,Y in R2``

    2. Extract scale components of ``A``


    :param pt: estimate transform around this point
    :param r:  radius around the point (default is 1)
    :param tr: Point transforming function ``Sequence[XY[float]] -> Sequence[XY[float]]``

    :return: ``(sx, sy)`` where ``sx > 0, sy > 0``
    """
    pts0 = [(0, 0), (-1, 0), (0, -1), (1, 0), (0, 1)]
    x0, y0 = pt.xy
    if r is None:
        XX = [xy_(float(x + x0), float(y + y0)) for x, y in pts0]
    else:
        XX = [xy_(float(x * r + x0), float(y * r + y0)) for x, y in pts0]
    YY = tr(XX)
    A = affine_from_pts(XX, YY)
    return get_scale_from_linear_transform(A)


def _same_crs_pix_transform(src: GeoBox, dst: GeoBox) -> LinearPointTransform:
    assert src.crs == dst.crs
    _fwd = (~dst.transform) * src.transform  # src -> dst
    return LinearPointTransform(_fwd)


def compute_axis_overlap(
    Ns: int, Nd: int, s: float, t: float
) -> Tuple[NormalizedSlice, NormalizedSlice]:
    """
    s, t define linear transform from destination coordinate space to source
    >>  x_s = s * x_d + t

    Ns -- number of pixels along some dimension of source image: (0, Ns)
    Nd -- same as Ns but for destination image

    :returns: (slice in the source image,
               slice in the destination image)
    """

    needs_flip = s < 0

    if needs_flip:
        # change s, t to map into flipped src, i.e. src[::-1]
        s, t = -s, Ns - t

    assert s > 0

    # x_d = (x_s - t)/s => 1/s * x_s + t*(-1/s)
    #
    # x_d = s_ * x_s + t_
    s_ = 1.0 / s
    t_ = -t * s_

    if t < 0:
        #  |<------- ... D
        #      |<--- ... S
        _in = (0, min(math.floor(t_), Nd))
    else:
        #        |<--... D
        # |<---------... S
        _in = (min(math.floor(t), Ns), 0)

    a = math.ceil(Nd * s + t)
    if a <= Ns:
        # ...----->|    D
        # ...-------->| S
        _out = (max(a, 0), Nd)
    else:
        # ...-------->|  D
        # ...----->|     S
        _out = (Ns, max(0, math.ceil(Ns * s_ + t_)))

    src, dst = (slice(_in[i], _out[i]) for i in range(2))

    if needs_flip:
        # remap src from flipped space to normal
        src = slice(Ns - src.stop, Ns - src.start)  # type: ignore

    return (src, dst)


def box_overlap(
    src_shape: SomeShape, dst_shape: SomeShape, ST: Affine
) -> Tuple[NormalizedROI, NormalizedROI]:
    """
    Compute overlap between two image planes.

    Given two image planes whose coordinate systems are related via scale and
    translation only, find overlapping regions within both.

    :param src_shape:
      Shape of source image plane

    :param dst_shape:
      Shape of destination image plane

    :param ST:
      Affine transform with only scale/translation, direction is: ``Xsrc = ST*Xdst``

    :returns: ``(src_roi, dst_roi)``
    """

    src_shape = shape_(src_shape)
    dst_shape = shape_(dst_shape)

    (sx, _, tx, _, sy, ty, *_) = ST

    s0, d0 = compute_axis_overlap(src_shape.y, dst_shape.y, sy, ty)
    s1, d1 = compute_axis_overlap(src_shape.x, dst_shape.x, sx, tx)
    return (s0, s1), (d0, d1)


def native_pix_transform(src: GeoBoxBase, dst: GeoBoxBase) -> PointTransform:
    """

    direction: from src to dst
    .back: goes the other way
    .linear: None|Affine linear transform src->dst if transform is linear (i.e. same CRS)
    """
    # Special case CRS_in == CRS_out
    if isinstance(src, GeoBox) and isinstance(dst, GeoBox) and src.crs == dst.crs:
        return _same_crs_pix_transform(src, dst)

    return GbxPointTransform(src, dst)


def _pick_read_scale(scale: float, tol: float = 1e-3) -> int:
    assert scale > 0
    # scale < 1 --> 1
    # Else: scale down to nearest integer, unless we can scale up by less than tol
    #
    # 2.999999 -> 3
    # 2.8 -> 2
    # 0.3 -> 1

    if scale < 1:
        return 1

    # if close to int from below snap to it
    scale = maybe_int(scale, tol)

    # otherwise snap to nearest integer below
    return int(scale)


def _can_paste(
    A: Affine, stol: float = 1e-3, ttol: float = 1e-2
) -> Tuple[bool, Optional[str]]:
    """
    Check if can read (possibly with scale) and paste, or do we need to read then reproject.

    :param A: Coordinate mapping from dst to src ``X_src = A*X_dst``

    :returns: (True, None) if one can just read and paste
    :returns: (False, Reason) if pasting is not possible, so need to reproject after reading
    """

    if not is_affine_st(A):
        return (False, "has rotation or shear")

    sx, sy = get_scale_from_linear_transform(A).xy
    scale = min(sx, sy)

    if not is_almost_int(scale, stol):  # non-integer scaling
        return False, "non-integer scale"

    read_scale = _pick_read_scale(scale)
    # A_ maps coords from `dst` to `src.overview[scale]`
    A_ = Affine.scale(1 / read_scale, 1 / read_scale) * A

    (sx, _, tx, _, sy, ty, *_) = A_  # tx, ty are in dst pixel space

    # Expect identity for scale change
    if any(abs(abs(s) - 1) > stol for s in (sx, sy)):  # not equal scaling across axis?
        return False, "sx!=sy, probably"

    # Check if sub-pixel translation within bounds
    if not all(is_almost_int(t, ttol) for t in (tx, ty)):
        return False, "sub-pixel translation"

    return True, None


def _relative_rois(
    src: GeoBoxBase,
    dst: GeoBoxBase,
    tr: PointTransform,
    pts_per_side: int,
    padding: int,
    align: Optional[int],
):
    _XY = tr.back(unstack_xy(gbox_boundary(dst, pts_per_side)))
    roi_src = roi_from_points(stack_xy(_XY), src.shape, padding, align=align)

    if roi_is_empty(roi_src):
        return (roi_src, np.s_[0:0, 0:0])

    # project src roi back into dst and compute roi from that
    xy = tr(unstack_xy(roi_boundary(roi_src, pts_per_side)))

    # `padding=0` is to avoid adding padding twice
    roi_dst = roi_from_points(stack_xy(xy), dst.shape, padding=0)
    return (roi_src, roi_dst)


def compute_reproject_roi(
    src: GeoBoxBase,
    dst: GeoBoxBase,
    ttol: float = 0.05,
    stol: float = 1e-3,
    padding: Optional[int] = None,
    align: Optional[int] = None,
) -> ReprojectInfo:
    """
    Compute reprojection information.

    Given two GeoBoxes find the region within the source GeoBox that overlaps
    with the destination GeoBox, and also compute the scale factor (>1 means
    shrink). Scale is chosen such that if you apply it to the source image
    before reprojecting, then reproject will have roughly no scale component.

    So we are breaking up reprojection into two stages:

    1. Scale in the native pixel CRS
    2. Reprojection (possibly non-linear with CRS change)

    .. code-block:: text

       - src[roi] -> scale      -> reproject -> dst  (using native pixels)
       - src(scale)[roi(scale)] -> reproject -> dst  (using overview image)

    Here ``roi`` is "minimal", padding is configurable though, so you only read what you need.
    Also scale can be used to pick the right kind of overview level to read.

    Applying reprojection in two steps allows us to use pre-computed overviews,
    particularly useful when shrink factor is large. But even for data sources
    without overviews there are advantages for shrinking source image before
    applying reprojection: mainly quality of the output (reduces aliasing for
    large shrink factors), improved efficiency of the computation is likely as
    well.

    Also compute and return ROI of the dst geobox that is affected by src.

    If padding is ``None`` "appropriate" padding will be used depending on the
    transform between ``src<>dst``:

    * No padding beyond sub-pixel alignment if Scale+Translation

    * 1 pixel source padding in all other cases

    :param src:
      Geobox of the source image

    :param dst:
      Geobox of the destination image

    :param padding:
      Optional padding in source pixels

    :param align:
      Optional pixel alignment in pixels, used on both source and destination.

    :param tol:
      Sub-pixel translation tolerance as pixel fraction.

    :param stol:
      Scale tolerance for pasting

    :returns:
      An instance of ``ReprojectInfo`` class.
    """
    # pylint: disable=too-many-locals

    pts_per_side = 5

    tr = native_pix_transform(src, dst)

    if tr.linear is None:
        padding = 1 if padding is None else padding
        roi_src, roi_dst = _relative_rois(
            src, dst, tr, pts_per_side=pts_per_side, padding=padding, align=align
        )

        if not roi_is_empty(roi_dst):
            center_pt = xy_(roi_center(roi_dst)[::-1])
            scale2 = get_scale_at_point(center_pt, tr.back)
            scale = min(scale2.xy)
            read_shrink = _pick_read_scale(scale)
        else:
            scale = 0
            scale2 = XY(x=0, y=0)
            read_shrink = 1

        return ReprojectInfo(
            roi_src=roi_src,
            roi_dst=roi_dst,
            scale=scale,
            scale2=scale2,
            paste_ok=False,
            read_shrink=read_shrink,
            transform=tr,
        )

    # Same projection case
    #
    A = tr.back.linear  # dst->src
    scale2 = get_scale_from_linear_transform(A)
    scale = min(scale2.xy)
    read_shrink = _pick_read_scale(scale)

    paste_ok = False
    tight_ok = align in (None, 0) and padding in (0, None)

    if tight_ok:
        paste_ok, _ = _can_paste(A, ttol=ttol, stol=stol)

    if paste_ok:
        if read_shrink == 1:
            A_ = snap_affine(A, ttol=ttol, stol=stol)
            roi_src, roi_dst = box_overlap(src.shape, dst.shape, A_)
        else:
            # compute overlap in scaled down image, then upscale source overlap
            assert isinstance(src, GeoBox)
            _src = src.zoom_out(read_shrink)
            A_ = snap_affine(Affine.scale(1 / read_shrink) * A, ttol=ttol, stol=stol)
            roi_src, roi_dst = box_overlap(_src.shape, dst.shape, A_)
            roi_src = scaled_up_roi(roi_src, read_shrink)
    else:
        padding = 1 if padding is None else padding
        roi_src, roi_dst = _relative_rois(
            src, dst, tr, pts_per_side=2, padding=padding, align=align
        )

    return ReprojectInfo(
        roi_src=roi_src,
        roi_dst=roi_dst,
        scale=scale,
        scale2=scale2,
        paste_ok=paste_ok,
        read_shrink=read_shrink,
        transform=tr,
    )


def compute_output_geobox(
    gbox: Union[GeoBox, GCPGeoBox],
    crs: SomeCRS,
    *,
    resolution: Union[SomeResolution, Literal["auto", "fit", "same"]] = "auto",
    shape: Union[SomeShape, int, None] = None,
    tight: bool = False,
    anchor: GeoboxAnchor = "default",
    tol: float = 0.01,
    round_resolution: Union[None, bool, Callable[[float, str], float]] = None,
) -> GeoBox:
    """
    Compute output ``GeoBox``.

    Find best fitting, axis aligned GeoBox in a different coordinate reference
    given source ``GeoBox`` on input.

    :param gbox:
       Source geobox.

    :param crs:
       Desired CRS of the output

    :param resolution:

       * "same" use exactly the same resolution as src
       * "fit" use center pixel to determine scale change between the two
       * | "auto" is to use the same resolution on the output if CRS units are
         | the same between the source and destination and otherwise use "fit"
       * Ignored if ``shape=`` is supplied
       * Else resolution in the units of the output crs

    :param shape:
      Span that many pixels, if it's a single number then span that many pixels
      along the longest dimension, other dimension will be computed to maintain
      roughly square pixels. Takes precedence over ``resolution=`` parameter.

    :param tight:
      By default output pixel grid is adjusted to align pixel edges to X/Y axis,
      suppling ``tight=True`` produces unaligned geobox on the output.

    :param anchor:
      Control pixel snapping, default is to snap pixel edge to ``X=0,Y=0``.
      Ignored when ``tight=True`` is supplied.

    :param tol:
       Fraction of the output pixel that can be ignored, defaults to 1/100.
       Bounding box of the output geobox is allowed to be smaller by that amount
       than transformed footprint of the original.

    :param round_resolution:
      ``round_resolution(res: float, units: str) -> float``

    :return:
       Similar resolution, axis aligned geobox that fully encloses source one
       but in a different projection.
    """
    # pylint: disable=too-many-locals,too-many-arguments
    src_crs = gbox.crs
    assert src_crs is not None

    # figure out "true" dts_crs (handles "utm" -> actual CRS)
    bbox = gbox.footprint(crs, buffer=0.9, npoints=100).boundingbox
    dst_crs = bbox.crs
    assert dst_crs is not None

    if (
        dst_crs == src_crs
        and resolution in ("auto", "same")
        and shape is None
        and anchor == "default"
        and isinstance(gbox, GeoBox)
    ):
        return gbox

    same_units = src_crs.units == dst_crs.units

    if shape is not None:
        res: Optional[SomeResolution] = None
    elif resolution == "same":
        res = gbox.resolution
    elif resolution == "auto" and same_units:
        res = gbox.resolution
    elif resolution in ("fit", "auto"):
        # get initial resolution by computing 1x1 bounding box of the center pixel
        #
        cp = gbox.center_pixel
        cp_bbox = cp.extent.to_crs(dst_crs).boundingbox
        dst_ = GeoBox.from_bbox(cp_bbox, dst_crs, shape=(1, 1), tight=True)

        # further adjust that via fitting
        #  Y = tr(X) => Y ~= s*X + t
        # where X is in `dst_` and Y is in `gbox`
        # .. so a better fit resolution is `_dst.res / s`
        # .. but we want square pixels, so pick average between x and y
        sx, sy = get_scale_at_point(xy_(0.5, 0.5), native_pix_transform(dst_, cp)).xy

        # always produces square pixels on output with inverted Y axis
        avg_res = (abs(dst_.resolution.x / sx) + abs(dst_.resolution.y / sy)) / 2
        if round_resolution is not None:
            if isinstance(round_resolution, bool):
                if round_resolution:
                    avg_res = round(avg_res, 0)
            else:
                avg_res = round_resolution(avg_res, dst_crs.units[0])
        res = res_(avg_res)
    else:
        if isinstance(resolution, str):
            raise ValueError(
                f"Resolution ought to be one of: same,auto,fit, not '{resolution}'"
            )

        res = res_(resolution)

    return GeoBox.from_bbox(
        bbox,
        dst_crs,
        shape=shape,
        resolution=res,
        tight=tight,
        anchor=anchor,
        tol=tol,
    )



from __future__ import annotations

import pickle
from pathlib import Path
from uuid import uuid4

import pytest

_ = pytest.importorskip("odc.geo.cog")

from odc.geo.cog._mpu_fs import MPUFileSink

# pylint: disable=protected-access


def slurp(path: Path) -> bytes:
    with path.open("rb") as f:
        return f.read()


@pytest.mark.parametrize("parts_base", [f"parts-3782781", None])
@pytest.mark.parametrize("num_parts", [1, 2, 3, 10])
def test_filesink(tmp_path: Path, parts_base: str | None, num_parts: int):
    dst = Path(tmp_path / f"{uuid4().hex}.bin")
    assert dst.exists() is False
    if parts_base is not None:
        parts_base = str(tmp_path / parts_base)

    sink = MPUFileSink(dst, parts_base)
    assert sink.__dask_tokenize__() == sink.__dask_tokenize__()
    assert sink.max_part > sink.min_part
    assert sink.max_write_sz > sink.min_write_sz

    # simulate distributed writing
    sink2 = pickle.loads(pickle.dumps(sink))

    chunks = [
        (idx, f"part: {idx}\n...{idx}\n".encode("utf8"))
        for idx in range(1, num_parts + 1)
    ]
    expected_content = b"".join(data for _, data in chunks)

    parts = []
    for idx, data in chunks:
        _sink = sink if idx % 2 == 0 else sink2
        parts.append(_sink(idx, data))

    for (idx, data), part in zip(chunks, parts):
        assert part["PartNumber"] == idx
        assert part["Size"] == len(data)
        part_path = Path(part["Path"])
        assert part_path.exists()
        assert slurp(part_path) == data

    assert sink._parts_dir.exists()
    assert dst.exists() is False

    assert sink.finalise(parts) == dst

    assert dst.exists()
    assert sink._parts_dir.exists() is False
    assert any(Path(p["Path"]).exists() for p in parts) is False

    assert slurp(dst) == expected_content



import base64
import warnings
from typing import Optional, Tuple, Union

import numpy as np
import rasterio
import rasterio.env
import rasterio.session
import xarray as xr

from ._interop import is_dask_collection
from ._rgba import replace_transparent_pixels

# Driver, Compress Option Name
_fmt_info = {
    "png": ("PNG", "zlevel", "image/png"),
    "jpeg": ("JPEG", "quality", "image/jpeg"),
    "webp": ("WEBP", "quality", "image/webp"),
}


def _verify_can_compress(xx: xr.DataArray):
    """Returns error if array dimensions are not suitable for compress"""
    if xx.ndim > 2:
        xx = xx.squeeze()

    if xx.ndim not in (2, 3):
        raise ValueError(
            f"Expected a 2 or 3 dimensional array; got {xx.ndim} dimensions {xx.dims}."
        )


def _compress_image(im: np.ndarray, driver="PNG", **opts) -> bytes:
    if im.ndim > 2:
        im = np.squeeze(im)

    if im.ndim == 3:
        h, w, nc = im.shape  # type: ignore
        bands = np.transpose(im, axes=(2, 0, 1))  # Y,X,B -> B,Y,X
    elif im.ndim == 2:
        (h, w), nc = im.shape, 1
        bands = im.reshape(nc, h, w)
    else:
        raise ValueError(
            f"Expected a 2 or 3 dimensional array; got {im.ndim} dimensions."
        )

    rio_opts = {
        "width": w,
        "height": h,
        "count": nc,
        "driver": driver,
        "dtype": im.dtype,
        **opts,
    }
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", rasterio.errors.NotGeoreferencedWarning)

        with rasterio.env.Env(session=rasterio.session.DummySession()):
            with rasterio.MemoryFile() as mem:
                with mem.open(**rio_opts) as dst:
                    dst.write(bands)
                return mem.read()


def compress(
    xx,
    /,
    *args,
    as_data_url=False,
    transparent: Optional[Tuple[int, int, int]] = None,
    **kw,
) -> Union[str, bytes]:
    """
    Save image to RAM in jpeg/png/webp format.

    .. code:: python

       png_bytes = compress(xx)  # default is PNG
       png_bytes = compress(xx, "png", 9) # compression settings

       # - Make data url with JPEG quality of 85
       # - Use black for transparent pixels (JPEG doesn't support transparency)
       # - Result is an ASCII string
       url = compress(xx, "jpeg", 85, as_data_url=True, transparent=(0,0,0))

    :param xx: DataArray to compress
    :param transparent:
      Pixel value to use for transparent pixels, useful for jpeg output.

    """
    # Raise error early if xx has unsuitable dims
    _verify_can_compress(xx)

    fmt = "png"
    opts = {}
    if len(args) >= 1:
        fmt, *_ = args

    driver, opt_name, mime_type = _fmt_info.get(fmt.lower(), (None, "", ""))
    if driver is None:
        raise ValueError(f"Format '{fmt}', is not suported, try jpeg/png/webp")

    if len(args) >= 2:
        _, comp, *_ = args
        opts[opt_name] = comp

    if isinstance(xx, xr.DataArray):
        xx = xx.data

    if is_dask_collection(xx):
        xx = xx.compute()

    if transparent is not None:
        xx = replace_transparent_pixels(xx, transparent)

    _bytes = _compress_image(xx, driver, **opts, **kw)
    if not as_data_url:
        return _bytes

    return f"data:{mime_type};base64," + base64.encodebytes(_bytes).decode("ascii")
